# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# -> Notes: line 69 explanation missing

from moleculekit.molecule import Molecule
from moleculekit.readers import _getPDB
import numpy as np
import copy
from tqdm import tqdm
import logging
import unittest
logger = logging.getLogger(__name__)


def _get_cellloc(center, angles, box_size):
    """
    Computes the Unit Cell's location of the molecule object.
    Takes as argument the Euler angles of the Unit Cell's sides and performs
    a small correction to account for oblique sides.
    Allows to correct x position dependent on y and z.
    """
    change = np.zeros(3)
    change[0] = center[1] / np.tan(angles[2]) + center[2] / np.tan(
        angles[1])  # get min_value x unit cell at given y,z
    return np.floor((center - change) / box_size)  # correct x axis boundaries according to angles


def _is_outside(center, cellloc):
    """
    Checks if a molecule object is inside the defined unit cell.
    Takes as arguments the size vector [numpy array with 3 coordinates]
    and a vector [numpy array with three floats] with the Euler angles of
    the Unit Cell.
    Modifies the attribute:
           -> center: mean center of the molecule object.

    Creates the attributes:
           -> cellloc: location of the molecule object within the Unit Cell.
    Returns True if the molecule object is outside the Unit Cell, else
    returns False.
    """
    return np.sum(np.abs(cellloc)) > 0, cellloc, center


def _move_inside(mol, axes, cellloc, center):
    """
    Translates an HTMD molecule object inside another Unit Cell.
    Takes as argument the vectors defining the Unit Cell the molecule
    object is currently in and translates it inside another Unit Cell.
    """
    neworigin = np.multiply(axes.transpose(), cellloc).transpose()  # computes translation to be applied to copy
    neworigin = np.sum(neworigin, axis=0)
    newcenter = center - neworigin  # computes difference between copy center and origin of its unit cell
    mol.center(loc=newcenter)  # places copy in the target unit cell (at an equivalent position to that of original unit cell)


def _place_crystal(mol, size, angles, axes):
    """
    Places MoleculeCopy object inside crystal.
    Uses methods 'is_outside' and 'move_inside'.
    """
    center = np.mean(mol.get('coords'), axis=0)
    cellloc = _get_cellloc(center, angles, size)  # integer from division for each axis = unit cell identifier
    if _is_outside(center, cellloc):
        _move_inside(mol, axes, cellloc, center)


def generateCrystalPacking(pdbid, hexagonal=False, visualize=False, viewerhandle=None):
    """
    Generates the crystal packing of a PDB protein.

    It is possible to inspect it immediately with the visualize option. It can only be generated if there is
    crystallographic information in the PDB entry.

    Parameters
    ----------
    pdbid : str
        ID from the Protein Databank
    hexagonal : bool
        # TODO: Undocumented
    visualize : bool
        If True, this function also visualizes the crystal packing
    viewerhandle : :class:`VMD <moleculekit.vmdviewer.VMD>` object, optional
        A specific viewer in which to visualize the molecule. If None it will use the current default viewer.
    Returns
    -------
    mol : :class:`Molecule`
        Molecule object with the crystal packing
    """

    if not isinstance(pdbid, str):
        raise ValueError('pdbid should be a string')

    filename, _ = _getPDB(pdbid)

    mol = Molecule(filename)

    if viewerhandle is not None and not visualize:
        logger.warning('A viewerhandle was passed. Setting visualize to True.')
        visualize = True
    if visualize and viewerhandle is None:
        from moleculekit.vmdviewer import getCurrentViewer
        viewerhandle = getCurrentViewer()

    if mol.crystalinfo is None or 'numcopies' not in mol.crystalinfo:
        raise RuntimeError('No crystallography data found in Molecule.')
    ci = mol.crystalinfo

    alpha, beta, gamma, a, b, c = ci['alpha'], ci['beta'], ci['gamma'], ci['a'], ci['b'], ci['c']
    alpha = np.deg2rad(float(alpha))
    beta = np.deg2rad(float(beta))
    gamma = np.deg2rad(float(gamma))

    caux = (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma)
    axes = np.array([[a, 0, 0], [b * np.cos(gamma), b * np.sin(gamma), 0], [c * np.cos(beta), c * caux,
                    c * np.sqrt(1 - np.cos(beta) ** 2 - caux ** 2)]])
    size = np.array([axes[0][0], axes[1][1], axes[2][2]])

    molunit = Molecule()

    # Creates copies of the molecule and places them correctly inside the complete Unit Cell
    hexagonal_molunit = None
    for i in tqdm(range(ci['numcopies']), desc='Generating symmetry mates'):
        molecule = mol.copy()
        molecule.segid[:] = str(i+1)
        # apply SMTRY (Crystal Symmetry) operations
        molecule.rotateBy(ci['rotations'][i])
        molecule.moveBy(ci['translations'][i])
        # apply translation to inside of same Unit Cell.
        _place_crystal(molecule, size, [alpha, beta, gamma], axes)
        # pack copies to target Unit Cell
        molunit.append(molecule)
    if ci['sGroup'][0] == 'H' and hexagonal:
        hexagonal_molunit = Molecule()
        _build_hexagon(molunit, hexagonal_molunit)

    if hexagonal_molunit is not None:
        if visualize:
            hexagonal_molunit.view(style='NewCartoon', viewerhandle=viewerhandle)
        else:
            return hexagonal_molunit
    else:
        if visualize:
            molunit.view(style='NewCartoon', viewerhandle=viewerhandle)
        else:
            return molunit

    if visualize:
        _draw_cell(axes, ci['sGroup'], viewerhandle, hexagonal=hexagonal)


def _draw_cell(axes, group, viewerhandle, hexagonal=False):
    """
    Draws lines in the viewer that represent Unit Cell.

    Parameters
    ----------
    hexagonal : bool
        If True, draws the Unit Cell with hexagonal boundaries if it is a hexagon. Else it draws a rectangle or
        trapezoid.
    """
    baseorigin = np.array([0, 0, 0])
    base_lo_l = axes[0]
    base_up_r = axes[1]
    base_lo_r = axes[0] + axes[1]
    top_origin = axes[2]
    top_up_r = axes[1] + axes[2]
    top_lo_l = axes[0] + axes[2]
    top_lo_r = axes.sum(axis=0)
    points = np.vstack((baseorigin, base_lo_l, base_up_r, base_lo_r))
    points = np.vstack((points, top_origin, top_lo_l, top_up_r, top_lo_r))
    if group[0] == 'H' and hexagonal:
        # Lattice is hexagonal, can be constructed by applying two 120º
        # rotations over the z axis
        alpha = np.deg2rad(120)
        rot_mat = np.array([[np.cos(alpha), -np.sin(alpha), 0],
                            [np.sin(alpha), np.cos(alpha), 0],
                            [0, 0, 1]])
        rotpoints1 = np.dot(rot_mat, points.transpose())
        rotpoints2 = np.dot(rot_mat, rotpoints1)
        points = np.vstack((points, rotpoints1.transpose(),
                            rotpoints2.transpose()))
        lines_draw = np.array([[1, 3], [1, 5], [2, 3], [2, 6], [3, 7],
                               [5, 7], [6, 7]])
        lines_draw = np.vstack((lines_draw, lines_draw + 8, lines_draw + 16))
    else:
        lines_draw = [[0, 1], [0, 2], [0, 4], [1, 3], [1, 5], [2, 3],
                      [2, 6], [3, 7], [4, 5], [4, 6], [5, 7], [6, 7]]
    for row in lines_draw:
        viewerhandle.send('draw line {{{0} {1} {2}}} {{{3} {4} {5}}}'.format(
            *(tuple(points[row[0]]) + tuple(points[row[1]]))))


def _build_hexagon(molunit, hexagonal_molunit):
    """
    Draws the copies within the Unit Cell stored in molunit as a hexagon by performing three consecutive rotations
    on the Z-axis with an angle of 120 degrees and places them in a new Molecule Object.
    Creates a new attribute, hexagonal_molunit, which stores the created Unit Cell HTMD molecule object.
    """

    for i in range(1, 4):
        cellunit = copy.deepcopy(molunit)  # TODO: This cant possibly work
        alpha = np.deg2rad(120 * i)
        rot_mat = np.array([[np.cos(alpha), -np.sin(alpha), 0],
                            [np.sin(alpha), np.cos(alpha), 0],
                            [0, 0, 1]])
        cellunit.rotateBy(rot_mat)
        hexagonal_molunit.append(cellunit)


class _TestCrystalPacking(unittest.TestCase):
    def test_crystalpacking_visualization(self):
        from moleculekit.vmdviewer import getCurrentViewer

        viewer = getCurrentViewer(dispdev='text')
        generateCrystalPacking('3ptb', visualize=True, viewerhandle=viewer)

    def test_crystalpacking_asymmetric_unit(self):
        from moleculekit.home import home
        from os.path import join
        from moleculekit.molecule import mol_equal

        mol = generateCrystalPacking('2hhb')
        refmol = Molecule(join(home(dataDir='test-crystalpacking'), '2hhb_packing.pdb'))
        assert mol_equal(mol, refmol, fieldPrecision={'coords': 1e-3})

    def test_crystalpacking_biological_unit(self):
        from moleculekit.home import home
        from os.path import join
        from moleculekit.molecule import mol_equal

        mol = generateCrystalPacking('1out')
        refmol = Molecule(join(home(dataDir='test-crystalpacking'), '1out_packing.pdb'))
        assert mol_equal(mol, refmol, fieldPrecision={'coords': 1e-3})


if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import abc


class Projection(abc.ABC):
    """
    Parent class for all trajectory projecting classes. Defines abstract functions.
    """
    def __init__(self):
        self._cache = {}

    @abc.abstractmethod
    def project(self, mol):
        """ Subclasses need to implement and overload this method """
        return

    @abc.abstractmethod
    def getMapping(self, mol):
        return

    @abc.abstractmethod
    def _calculateMolProp(self, mol, props='all'):
        return

    def _setCache(self, mol):
        resdict = self._calculateMolProp(mol)
        self._cache.update(resdict)

    def _getMolProp(self, mol, prop):
        if prop in self._cache:
            resdict = self._cache
        else:
            resdict = self._calculateMolProp(mol, [prop] if prop != 'all' else 'all')
            
        if prop == 'all':
            return resdict
        else:
            return resdict[prop]

    def copy(self):
        """ Produces a deep copy of the object
        """
        from copy import deepcopy
        return deepcopy(self)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricSphericalCoordinate(Projection):
    """ Creates a MetricSphericalCoordinate object that calculates the spherical coordinates between two centers of
    masses from a set of trajectories.

    Parameters
    ----------
    refmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we will align.
    targetcom : str
        Atom selection string from which to calculate the target center of mass.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    refcom : str
        Atom selection string from which to calculate the reference center of mass.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    trajalnsel : str, optional
        Atom selection string for the trajectories from which to align to the reference structure.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    refalnsel : str, optional
        Atom selection string for `refmol` from which to align to the reference structure. If None, it defaults to the
        same as `trajalnstr`. See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    centersel : str, optional
        Atom selection string around which to wrap the simulation.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    pbc : bool
        Enable or disable coordinate wrapping based on periodic boundary conditions.

    Returns
    -------
    metr : MetricCoordinate object
    """
    def __init__(self, refmol, targetcom, refcom, trajalnsel='protein and name CA', refalnsel=None, centersel='protein', pbc=True):
        super().__init__()

        if refalnsel is None:
            refalnsel = trajalnsel
        self._refmol = refmol
        self._refalnsel = self._refmol.atomselect(refalnsel)
        self._trajalnsel = trajalnsel
        self._centersel = centersel
        self._targetcom = targetcom
        self._refcom = refcom
        self._pbc = pbc

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        mol = mol.copy()
        if self._pbc:
            mol.wrap(self._centersel)
        mol.align(getMolProp('trajalnsel'), refmol=self._refmol, refsel=self._refalnsel)

        refcom = np.mean(mol.coords[getMolProp('refcom'), :, :], axis=0)
        targetcom = np.mean(mol.coords[getMolProp('targetcom'), :, :], axis=0)
        xyz = targetcom - refcom

        r = np.sqrt(np.sum(xyz ** 2, axis=0))
        theta = np.arccos(xyz[2, :] / r)
        phi = np.arctan2(xyz[1, :], xyz[0, :])

        return np.stack((r, theta, phi), axis=1)

    def _calculateMolProp(self, mol, props='all'):
        props = ('trajalnsel', 'targetcom', 'refcom', 'centersel') if props == 'all' else props
        res = {}

        if 'trajalnsel' in props:
            res['trajalnsel'] = mol.atomselect(self._trajalnsel)
            if np.sum(res['trajalnsel']) == 0:
                raise NameError('Alignment selection resulted in 0 atoms.')
        if 'targetcom' in props:
            res['targetcom'] = mol.atomselect(self._targetcom)
            if np.sum(res['targetcom']) == 0:
                raise NameError('Atom selection for `targetcom` resulted in 0 atoms.')
        if 'refcom' in props:
            res['refcom'] = mol.atomselect(self._refcom)
            if np.sum(res['refcom']) == 0:
                raise NameError('Atom selection for `refcom` resulted in 0 atoms.')
        return res

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        targetatomidx = np.where(getMolProp('targetcom'))[0]
        refatomidx = np.where(getMolProp('refcom'))[0]
        from pandas import DataFrame
        types = ['r', 'theta', 'phi']
        indexes = [[targetatomidx, refatomidx]] * 3
        description = ['r', 'theta', 'phi']
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


import unittest
class _TestMetricSphericalCoordinate(unittest.TestCase):
    def test_metricsphericalcoordinate(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        ref = mol.copy()
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))
        mol.bonds = mol._guessBonds()

        res = MetricSphericalCoordinate(ref, 'resname MOL', 'within 8 of resid 98').project(mol)
        _ = MetricSphericalCoordinate(ref, 'resname MOL', 'within 8 of resid 98').getMapping(mol)

        ref_array = np.load(path.join(home(dataDir='test-projections'), 'metricsphericalcoordinate', 'res.npy'))
        assert np.allclose(res, ref_array, rtol=0, atol=1e-04)



if __name__ == "__main__":
    unittest.main(verbosity=2)




try:
    import rdkit
except ImportError as e:
    raise ImportError('{}. You are probably missing the rdkit package. Please install it to support rdkit integration '
                      'features'.format(e))
from rdkit import Chem
from rdkit.Chem import ChemicalFeatures
from rdkit import RDConfig
from rdkit.Geometry import Point3D
import os
import numpy as np


def _convertMoleculeToRDKitMol(mol):
    """
    Convert Molecule to RDKit Mol
    Note: sometimes RDKit fails just writing a warning message to stderr, rather than raising an exception.
    Parameters
    ----------
    mol: Molecule
        Molecule to convert
    Return
    ------
    results: RDKit Mol
        RDKit Mol object
    Examples
    --------
    >>> from htmd.home import home
    >>> from moleculekit.molecule import Molecule
    >>> molFile = os.path.join(home('test-charge'), 'H2O.mol2')
    >>> mol = Molecule(molFile)
    >>> type(_convertMoleculeToRDKitMol(mol))
    <class 'rdkit.Chem.rdchem.Mol'>
    """

    from rdkit.Chem import MolFromMol2File
    import tempfile
    import sys

    rdkit_mol = None
    with tempfile.TemporaryFile(mode='w+') as stderr:

        # Redirect stderr to a file
        temp_fileno = os.dup(sys.stderr.fileno())
        os.dup2(stderr.fileno(), sys.stderr.fileno()) # Change process stderr
        sys.stderr = stderr # Change Python stderr

        # Convert Molecule to RDKit Mol
        with tempfile.TemporaryDirectory() as tmpDir:
            filename = os.path.join(tmpDir, 'mol.mol2')
            mol.write(filename)
            rdkit_mol = MolFromMol2File(filename, removeHs=False)

        # Reset stderr
        os.dup2(temp_fileno, sys.__stderr__.fileno())
        os.close(temp_fileno)
        sys.stderr = sys.__stderr__

        # Read RDKit warnings
        stderr.flush()
        stderr.seek(0)
        message = stderr.read()

    if rdkit_mol is None:
        raise RuntimeError(f'RDKit error\n{message}\n')

    if rdkit_mol.GetNumAtoms() != mol.numAtoms:
        raise RuntimeError(f'RDKit error: incorrect number of atoms.\n{message}\n')

    return rdkit_mol


def StandardPDBResidueChirality(rdmol):
    for a in rdmol.GetAtoms():
        if a.GetChiralTag() != rdkit.Chem.rdchem.CHI_UNSPECIFIED:
            info = a.GetMonomerInfo()
            if info and (info.GetMonomerType() == rdkit.Chem.rdchem.AtomMonomerType.PDBRESIDUE) \
                and not info.GetIsHeteroAtom() and not StandardPDBChiralAtom(info.GetResidueName(), info.GetName()):
                a.SetChiralTag(rdkit.Chem.rdchem.CHI_UNSPECIFIED)
                # TODO: Missing CIPCode unsetting


def StandardPDBChiralAtom(resname, atomname):
    if resname == 'GLY':
        return False
    if resname == 'ILE' or resname == 'THR':
        return atomname == 'CA' or atomname == 'CB'
    if resname in ('ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'HIS', 'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER', 'TRP', 'TYR', 'VAL'):
        return atomname == 'CA'
    return False


def toRDKITmol(mol, protidx, sanitize=True, removeHs=False):
    # Taken from rdkit/Code/GraphMol/FileParsers/PDBParser.cpp
    conformer = Chem.Conformer(len(protidx))
    conformer.Set3D(True)
    conformer.SetId(0)
    rdmol = Chem.RWMol()
    atomlist = []
    for ii, i in enumerate(protidx):
        a = Chem.Atom(mol.element[i])
        a.SetFormalCharge(int(mol.charge[i]))
        info = Chem.AtomPDBResidueInfo(atomName=mol.name[i],
                                    serialNumber=int(mol.serial[i]),
                                    altLoc=mol.altloc[i],
                                    residueName=mol.resname[i],
                                    residueNumber=int(mol.resid[i]),
                                    chainId=mol.chain[i],
                                    insertionCode=mol.insertion[i],
                                    occupancy=float(mol.occupancy[i]),
                                    tempFactor=float(mol.beta[i]),
                                    isHeteroAtom=mol.record[i]=='HETATM')
        a.SetMonomerInfo(info)

        rdmol.AddAtom(a)
        atomlist.append(a)
        coor = [float(c) for c in mol.coords[i, :, mol.frame]]
        conformer.SetAtomPosition(ii, Point3D(coor[0], coor[1], coor[2]))  # Correct the atom idx
    rdmol.AddConformer(conformer)

    # Here I diverge from the C++ parser because you cannot instantiate Chem.Bond objects in python
    # I also don't take into account double/triple bonds etc since I don't think we actually store them in Molecule
    for b in mol._getBonds():
        if b[0] in protidx and b[1] in protidx:
            bond = rdmol.GetBondBetweenAtoms(int(b[0]), int(b[1]))
            if bond is None:
                rdmol.AddBond(int(np.where(protidx == b[0])[0]), int(np.where(protidx == b[1])[0]), Chem.BondType.SINGLE)

    # Proximitybonds I already did by using _getBonds which calls _guessBonds
    # TODO: Set PDB double bonds

    # Calculate explicit valence of atoms
    for a in atomlist:
        pass

    if sanitize:
        if removeHs:
            Chem.RemoveHs(rdmol)
        else:
            Chem.SanitizeMol(rdmol)
    else:
        rdmol.UpdatePropertyCache()

    # Set tetrahedral chirality from 3D co-ordinates
    Chem.AssignAtomChiralTagsFromStructure(rdmol)
    StandardPDBResidueChirality(rdmol)

    return rdmol


if __name__ == '__main__':
    from moleculekit.molecule import Molecule

    mol = Molecule('3PTB')
    mol.filter('resname BEN')
    res = _convertMoleculeToRDKitMol(mol)
    assert res is not None

    # # We don't use this yet. It's experimental. So no point in wasting time testing it.
    # protidx = mol.atomselect('protein', indexes=True)

    # rdmol = toRDKITmol(mol, protidx)

    # fdefName = os.path.join(RDConfig.RDDataDir,'BaseFeatures.fdef')
    # factory = ChemicalFeatures.BuildFeatureFactory(fdefName)
    # feats = factory.GetFeaturesForMol(rdmol)



import string
import numpy as np
import logging


logger = logging.getLogger(__name__)


def _getChainAlphabet(mode):
    if mode == 'numeric':
        chain_alphabet = list(string.digits)
    elif mode == 'alphabetic':
        chain_alphabet = list(string.ascii_uppercase + string.ascii_lowercase)
    elif mode == 'alphanumeric':
        chain_alphabet = list(string.digits + string.ascii_uppercase + string.ascii_lowercase)
    else:
        raise RuntimeError('Invalid mode {}. Use \'numeric\', \'alphabetic\' or \'alphanumeric\''.format(mode))
    return chain_alphabet


def autoSegment(mol, sel='all', basename='P', spatial=True, spatialgap=4.0, field="segid", mode='alphanumeric', _logger=True):
    """ Detects resid gaps in a selection and assigns incrementing segid to each fragment

    !!!WARNING!!! If you want to use atom selections like 'protein' or 'fragment',
    use this function on a Molecule containing only protein atoms, otherwise the protein selection can fail.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    spatial : bool
        Only considers a discontinuity in resid as a gap of the CA atoms have distance more than `spatialgap` Angstrom
    spatialgap : float
        The size of a spatial gap which validates a discontinuity (A)
    field : str
        Field to fix. Can be "segid" (default), "chain", or "both"
    mode : str
        If set to 'numeric' it will use numbers for segment IDs.
        If set to 'alphabetic' it will use letters for segment IDs.
        If set to 'alphanumeric' it will use both numbers and letters for segment IDs.

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment(mol,'chain B','P')
    """
    from moleculekit.util import sequenceID
    mol = mol.copy()

    idx = mol.atomselect(sel, indexes=True)
    rid = mol.resid[idx].copy()
    residiff = np.diff(rid)
    gappos = np.where((residiff != 1) & (residiff != 0))[0]  # Points to the index before the gap!

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    used_chains = set(mol.chain)
    chain_alphabet = _getChainAlphabet(mode)

    available_chains = [x for x in chain_alphabet if x not in used_chains]

    idxstartseg = [idx[0]] + idx[gappos + 1].tolist()
    idxendseg = idx[gappos].tolist() + [idx[-1]]

    mol.set('segid', basename, sel)

    if len(gappos) == 0:
        mol.set('segid', basename+chain_alphabet[0], sel)
        return mol

    if spatial:
        residbackup = mol.resid.copy()
        mol.set('resid', sequenceID(mol.resid))  # Assigning unique resids to be able to do the distance selection

        todelete = []
        i = 0
        for s, e in zip(idxstartseg[1:], idxendseg[:-1]):
            # Get the carbon alphas of both residues  ('coords', sel='resid "{}" "{}" and name CA'.format(mol.resid[e], mol.resid[s]))
            ca1coor = mol.coords[(mol.resid == mol.resid[e]) & (mol.name == 'CA')]
            ca2coor = mol.coords[(mol.resid == mol.resid[s]) & (mol.name == 'CA')]
            if len(ca1coor) and len(ca2coor):
                dist = np.sqrt(np.sum((ca1coor.squeeze() - ca2coor.squeeze()) ** 2))
                if dist < spatialgap:
                    todelete.append(i)
            i += 1
        todelete = np.array(todelete, dtype=int)
        # Join the non-real gaps into segments
        idxstartseg = np.delete(idxstartseg, todelete+1)
        idxendseg = np.delete(idxendseg, todelete)

        mol.set('resid', residbackup)  # Restoring the original resids

    i = 0
    for s, e in zip(idxstartseg, idxendseg):
        # Fixup segid
        if field in ['segid', 'both']:
            newsegid = basename + str(i)
            if np.any(mol.segid == newsegid):
                raise RuntimeError('Segid {} already exists in the molecule. Please choose different prefix.'.format(newsegid))
            if _logger: logger.info('Created segment {} between resid {} and {}.'.format(newsegid, mol.resid[s], mol.resid[e]))
            mol.segid[s:e+1] = newsegid
        # Fixup chain
        if field in ['chain', 'both']:
            newchainid = available_chains[i]
            if _logger: logger.info('Set chain {} between resid {} and {}.'.format(newchainid, mol.resid[s], mol.resid[e]))
            mol.chain[s:e+1] = newchainid

        i += 1

    return mol


def autoSegment2(mol, sel='(protein or resname ACE NME)', basename='P', fields=('segid',), residgaps=False, residgaptol=1, chaingaps=True, mode='alphanumeric', _logger=True):
    """ Detects bonded segments in a selection and assigns incrementing segid to each segment

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    fields : tuple of strings
        Field to fix. Can be "segid" (default) or any other Molecule field or combinations thereof.
    residgaps : bool
        Set to True to consider gaps in resids as structural gaps. Set to False to ignore resids
    residgaptol : int
        Above what resid difference is considered a gap. I.e. with residgaptol 1, 235-233 = 2 > 1 hence is a gap. We set
        default to 2 because in many PDBs single residues are missing in the proteins without any gaps.
    chaingaps : bool
        Set to True to consider changes in chains as structural gaps. Set to False to ignore chains
    mode : str
        If set to 'numeric' it will use numbers for segment IDs.
        If set to 'alphabetic' it will use letters for segment IDs.
        If set to 'alphanumeric' it will use both numbers and letters for segment IDs.

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment2(mol)
    """
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import connected_components

    if isinstance(fields, str):
        fields = (fields,)

    sel += ' and backbone or (resname NME ACE and name N C O CH3)'  # Looking for bonds only over the backbone of the protein
    idx = mol.atomselect(sel, indexes=True)  # Keep the original atom indexes to map from submol to mol
    submol = mol.copy()  # We filter out everything not on the backbone to calculate only those bonds
    submol.filter(sel, _logger=False)
    bonds = submol._getBonds()  # Calculate both file and guessed bonds

    if residgaps:
        # Remove bonds between residues without continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        bonds = bonds[bondresiddiff <= residgaptol, :]
    else:
        # Warning about bonds bonding non-continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        if _logger and np.any(bondresiddiff > 1):
            for i in np.where(bondresiddiff > residgaptol)[0]:
                logger.warning('Bonds found between resid gaps: resid {} and {}'.format(submol.resid[bonds[i, 0]],
                                                                                        submol.resid[bonds[i, 1]]))
    if chaingaps:
        # Remove bonds between residues without same chain
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        bonds = bonds[bondsamechain, :]
    else:
        # Warning about bonds bonding different chains
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        if _logger and np.any(bondsamechain == False):
            for i in np.where(bondsamechain == False)[0]:
                logger.warning('Bonds found between chain gaps: resid {}/{} and {}/{}'.format(submol.resid[bonds[i, 0]],
                                                                                              submol.chain[bonds[i, 0]],
                                                                                              submol.resid[bonds[i, 1]],
                                                                                              submol.chain[bonds[i, 1]]
                                                                                              ))

    # Calculate connected components using the bonds
    sparsemat = csr_matrix((np.ones(bonds.shape[0] * 2),  # Values
                            (np.hstack((bonds[:, 0], bonds[:, 1])),  # Rows
                             np.hstack((bonds[:, 1], bonds[:, 0])))), shape=[submol.numAtoms, submol.numAtoms])  # Columns
    numcomp, compidx = connected_components(sparsemat, directed=False)

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    used_chains = set(mol.chain)
    chain_alphabet = _getChainAlphabet(mode)
    available_chains = [x for x in chain_alphabet if x not in used_chains]

    mol = mol.copy()
    prevsegres = None
    for i in range(numcomp):  # For each connected component / segment
        segid = basename + str(i)
        backboneSegIdx = idx[compidx == i]  # The backbone atoms of the segment
        segres = mol.atomselect('same residue as index {}'.format(' '.join(map(str, backboneSegIdx)))) # Get whole residues

        # Warning about separating segments with continuous resids
        if _logger and i > 0 and (np.min(mol.resid[segres]) - np.max(mol.resid[prevsegres])) == 1:
            logger.warning('Separated segments {} and {}, despite continuous resids, due to lack of bonding.'.format(
                            basename + str(i-1), segid))

        # Add the new segment ID to all fields the user specified
        for f in fields:
            if f != 'chain':
                if np.any(mol.__dict__[f] == segid):
                    raise RuntimeError('Segid {} already exists in the molecule. Please choose different prefix.'.format(segid))
                mol.__dict__[f][segres] = segid  # Assign the segid to the correct atoms
            else:
                mol.__dict__[f][segres] = available_chains[i % len(available_chains)]
        if _logger:
            logger.info('Created segment {} between resid {} and {}.'.format(segid, np.min(mol.resid[segres]),
                                                                            np.max(mol.resid[segres])))
        prevsegres = segres  # Store old segment atom indexes for the warning about continuous resids

    return mol

import unittest
class _TestPreparation(unittest.TestCase):
    def test_autoSegment(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from os import path

        p = Molecule(path.join(home(dataDir='test-autosegment'), '4dkl.pdb'))
        p.filter('(chain B and protein) or water')
        p = autoSegment(p, 'protein', 'P')
        m = Molecule(path.join(home(dataDir='test-autosegment'), 'membrane.pdb'))
        print(np.unique(m.get('segid')))

        mol = Molecule(path.join(home(dataDir='test-autosegment'), '1ITG_clean.pdb'))
        ref = Molecule(path.join(home(dataDir='test-autosegment'), '1ITG.pdb'))
        mol = autoSegment(mol, sel='protein')
        assert np.all(mol.segid == ref.segid)

        mol = Molecule(path.join(home(dataDir='test-autosegment'), '3PTB_clean.pdb'))
        ref = Molecule(path.join(home(dataDir='test-autosegment'), '3PTB.pdb'))
        mol = autoSegment(mol, sel='protein')
        assert np.all(mol.segid == ref.segid)

if __name__ == "__main__":
    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# The molvs.TautomerCanonicalizer was refactored by:
#           * adding conjugate length score
#           * tautomers filtering based on the scores computed and a given threshold (default: 2)

# Two class were implemented:
#           * MolFragmenter: a class that handles the molecule fragmentations into conjugates fragments and then
#             compute the longest path for electron delocalization
#           * Fragment: a class that is simply an object where to store information of each fragment generated

from rdkit import Chem
from rdkit.Chem import rdchem
from rdkit.Chem.rdchem import BondType
from moleculekit.smallmol.smallmol import SmallMol
from moleculekit.smallmol.tools.molvstautomer import TAUTOMER_TRANSFORMS, TAUTOMER_SCORES, MAX_TAUTOMERS, pairwise, memoized_property
import numpy as np
import logging

logger = logging.getLogger(__name__)

AROMATIC = rdchem.BondType.AROMATIC
SINGLE = rdchem.BondType.SINGLE
DOUBLE = rdchem.BondType.DOUBLE


class MolFragmenter:
    """
    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol - the rdkit molecule object

    Attributes
    ----------

    mol: rdkit.Chem.rdchem.Mol - the rdkit molecule object
    fragments: list - list of Fragment objects
    paths: list - list of list containing fragment objects part of the path


    """
    # conjugate scores. 1pt for each double bond
    SCORE_dict = {'aromatic': 3, 'double': 1}

    def __init__(self, mol):
        if isinstance(mol, SmallMol):
            mol = mol._mol

        self.mol = mol
        self.fragments = []
        self.paths = []

    def fragment(self):
        """
        The method creates and stores Fragment objects for each bond in the molecule and based on the atom type
        """

        # storing all double bonds idences by using rdkit.GetBondType
        all_double = [b.GetIdx() for b in self.mol.GetBonds() if b.GetBondType() == DOUBLE]
        # storing all aromatic bonds idences by using rdkit.GetBondType
        all_aromatics = []
        for ring in self.mol.GetRingInfo().BondRings():
            if sum([1 for bidx in ring if self.mol.GetBondWithIdx(bidx).GetBondType() == AROMATIC]) == len(ring):
                all_aromatics.append(list(ring))
        # storing all others bonds type
        try:  # for aromatics
            all_others = [b.GetIdx() for b in self.mol.GetBonds() if
                          b.GetIdx() not in np.concatenate(all_aromatics) and b.GetIdx() not in all_double]
        except:
            all_others = [b.GetIdx() for b in self.mol.GetBonds() if b.GetIdx() not in all_double]

        # generating Fragment object for each bond in the molecules based on their types
        nameint = 1
        for f in all_double:
            fr = Fragment(str(nameint), 'double')
            bond = self.mol.GetBondWithIdx(f)
            atoms = [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]
            fr.atoms = atoms
            self.fragments.append(fr)
            nameint += 1
        for f in all_aromatics:
            fr = Fragment(str(nameint), 'aromatic')
            atoms = []
            for bidx in f:
                bond = self.mol.GetBondWithIdx(bidx)
                atoms.extend([bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()])
            fr.atoms = np.unique(atoms).tolist()
            self.fragments.append(fr)
            nameint += 1
        for f in all_others:
            fr = Fragment(str(nameint), 'single')
            bond = self.mol.GetBondWithIdx(f)
            atoms = [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]
            fr.atoms = atoms
            self.fragments.append(fr)
            nameint += 1

    def get_fragment(self, attr, value):
        """
        The function returns the Fragment objects based on the attribute and value provided

        Parameters
        ----------
        attr: Fragment.attributename -  the Fragment object attribute  (selector)
        value: Fragment.attributevalue - the Fragment object attribute value to match

        Returns
        -------
        fragments: list -  The list of Fragment objects matching the parameters provided

        """
        fragments = [f for f in self.fragments if getattr(f, attr) == value]
        return fragments

    def search_fragment_by_atom(self, atom_idx, fragmenttype=None):
        """
        The function returns the Fragment objects based on the atomIdx  and fragmenttype provided (default: all type)

        Parameters
        ----------
        atom_idx: int - The atom idx
        fragmenttype:  Fragment.ftype value - The fragment type to match

        Returns
        -------
        fragment: list - The list of Fragment objects matching the parameters provided

        """

        if fragmenttype is not None:
            fragments = self.get_fragment('ftype', fragmenttype)
        else:
            fragments = self.fragments

        fragment = [fr for fr in fragments if atom_idx in fr.atoms]

        return fragment

    def get_num_fragments(self):
        """
        Returns: int
                 Total number of fragments.
        """

        return len(self.fragments)

    def get_graphs(self):
        """
        The function inspect for each conjugate Fragment (double aromatic) all the possible pattern
        that will set as connections
        """

        for fr in self.get_fragment('ftype', 'double'):
            graphs = []
            for frsingle in self.get_fragment('ftype', 'single'):
                graphs.extend([frsingle for a in fr.atoms if a in frsingle.atoms])
            fr.connections = [fr.name for fr in graphs]
            fr.graphs = graphs

        for fr in self.get_fragment('ftype', 'aromatic'):
            graphs = []
            for frsingle in self.get_fragment('ftype', 'single'):
                graphs.extend([frsingle for a in fr.atoms if a in frsingle.atoms])
            for fraromatic in self.get_fragment('ftype', 'aromatic'):
                if fr != fraromatic:
                    if len([a for a in fr.atoms if a in fraromatic.atoms]) == 2:
                        graphs.append(fraromatic)

            fr.connections = [fr.name for fr in graphs]
            fr.graphs = graphs

    def compute_paths(self):
        """
        The function inpect and detect all the possible conjugate paths
        """

        fragments = self.get_fragment('ftype', 'aromatic') + self.get_fragment('ftype', 'double')
        for fr in fragments:
            path = [fr]
            for fr2 in fragments:
                if fr == fr2:
                    continue
            path.extend([fr2 for f in path if f.ftype == 'aromatic' and
                         fr2.ftype == 'aromatic' and
                         f.name in fr2.connections])
            path.extend([fr2 for f in path for conn in f.connections if conn in fr2.connections])
        self.paths.append(list(set(path)))


    def get_longest_path(self):
        """
        The function return the longest path

        Returns
        -------
        highest_path: list - The list of fragmaent that represents the longest path

        """
        highest_num_fragments = max([len(p) for p in self.paths])

        longest_path = [p for p in self.paths if len(p) == highest_num_fragments][0]

        return longest_path

    def get_depiction_atoms(self, path):
        """
        The function returns all the atom indeces that are part of a conjugate path

        Parameters
        ---------
        path: list - List of Fragment objects that are part of the same conjugate path

        Returns
        -------
        atoms: list - List of atoms indeces part of a cpnjugate path
        """

        atoms = [a for fr in path for a in fr.atoms]
        return atoms

    def get_conjugate_length(self, path):
        """
        The function returns the length of the conjugate path

        Parameters
        ---------
        path: list - List of Fragment objects that are part of the same conjugate path

        Returns
        -------
        length: int - The lenght of the conjugate path
        """
        length = 0
        for fr in path:
            length += self.SCORE_dict[fr.ftype]
        return length


class Fragment:
    """
    The Fragment class store information about the molecule fragment

    Parameters
    ----------
    name: int - The id of the fragment
    ftype: str - The fragment type (choices: single, aromatic, double)


    Attributes
    ----------
    name: int - The id of the fragment
    ftype: str - The fragment type (choices: single, aromatic, double)
    atoms: list - List of the atom indeces
    graphs: list - List of the Fragment.name of the the Fragment object connected with this Fragment
    connection: list - List of the Fragment.name of the the Fragment object connected with
                       this Fragment (Only double and aromatic.)

    """

    def __init__(self, name, ftype):
        self.name = name
        self.ftype = ftype
        self.atoms = None
        self.graphs = None  # Fragment obj
        self.connections = []

    def get_connection(self):
        """The function returns the connections
        Returns
        -------
        connections: list - List of the Fragment.name of the the Fragment object connected with this Fragment
                            (Only double and aromatic.)
        """
        return self.connections


class TautomerCanonicalizer:
    """ The molvs class refactores"""

    def __init__(self, transforms=TAUTOMER_TRANSFORMS, scores=TAUTOMER_SCORES, max_tautomers=MAX_TAUTOMERS):
        """

        :param transforms: A list of TautomerTransforms to use to enumerate tautomers.
        :param scores: A list of TautomerScores to use to choose the canonical tautomer.
        :param max_tautomers: The maximum number of tautomers to enumerate, a limit to prevent combinatorial explosion.
        """
        self.transforms = transforms
        self.scores = scores
        self.max_tautomers = max_tautomers

    def __call__(self, mol):
        """Calling a TautomerCanonicalizer instance like a function is the same as calling its canonicalize(mol) method."""
        return self.canonicalize(mol)

    def canonicalize(self, mol):
        """Return a canonical tautomer by enumerating and scoring all possible tautomers.

        Parameters
        ----------
        mol: :class:`Mol <Chem.rdchem.Mol-class.html>`
            The input molecule

        Returns
        -------
        tau : :class:`Mol <Chem.rdchem.Mol-class.html>`
            The canonical tautomer.
        """
        # TODO: Overload the mol parameter to pass a list of pre-enumerated tautomers
        tautomers = self._enumerate_tautomers(mol)
        if len(tautomers) == 1:
            return tautomers[0]
        # Calculate score for each tautomer
        highest = None
        for t in tautomers:
            smiles = Chem.MolToSmiles(t, isomericSmiles=True)
            logger.debug('Tautomer: %s', smiles)
            score = 0
            # Add aromatic ring scores
            ssr = Chem.GetSymmSSSR(t)
            for ring in ssr:
                btypes = {t.GetBondBetweenAtoms(*pair).GetBondType() for pair in pairwise(ring)}
                elements = {t.GetAtomWithIdx(idx).GetAtomicNum() for idx in ring}
                if btypes == {BondType.AROMATIC}:
                    logger.debug('Score +100 (aromatic ring)')
                    score += 100
                    if elements == {6}:
                        logger.debug('Score +150 (carbocyclic aromatic ring)')
                        score += 150
            # Add SMARTS scores
            for tscore in self.scores:
                for _ in t.GetSubstructMatches(tscore.smarts):
                    logger.debug('Score %+d (%s)', tscore.score, tscore.name)
                    score += tscore.score
            # Add (P,S,Se,Te)-H scores
            for atom in t.GetAtoms():
                if atom.GetAtomicNum() in {15, 16, 34, 52}:
                    hs = atom.GetTotalNumHs()
                    if hs:
                        logger.debug('Score %+d (%s-H bonds)', -hs, atom.GetSymbol())
                        score -= hs
            # Set as highest if score higher or if score equal and smiles comes first alphabetically
            if not highest or highest['score'] < score or (highest['score'] == score and smiles < highest['smiles']):
                logger.debug('New highest tautomer: %s (%s)', smiles, score)
                highest = {'smiles': smiles, 'tautomer': t, 'score': score}
        return highest['tautomer']

    @memoized_property
    def _enumerate_tautomers(self):
        from moleculekit.smallmol.tools.molvstautomer import TautomerEnumerator
        return TautomerEnumerator(self.transforms, self.max_tautomers)

    def depict_tautomers(self, tautomers, scores, atoms, details=None):
        import matplotlib.pyplot as plt
        from rdkit.Chem import Draw
        from math import ceil
        
        dep_for_row = len(tautomers) if len(tautomers) < 3 else 3
        legends_tmp = [s for s in scores]
        legends = []
        if details is not None:
            for n, l in enumerate(legends_tmp):
                l = 'TotalScore: {}\n {}'.format(l, "\n".join(['{}: {}'.format(k, v) for k, v in details[n].items()]))
                legends.append(l)

        fig = plt.figure(figsize=(10, 10), dpi=100, tight_layout=True)
        for n, (t, l, a) in enumerate(zip(tautomers, legends, atoms)):
            depcition = Draw.MolToImage(t, subImgSize=(300, 300), highlightAtoms=a)
            ax = fig.add_subplot(ceil(len(tautomers) / dep_for_row), dep_for_row, n + 1)
            ax.set_xlabel(l, fontsize=5)
            ax.imshow(depcition, aspect='equal')
        plt.show()

    def get_conjugate(self, tautomer):
        """The functions called the MolFragmeter and returns the conjugate length  and the atoms member of the path

        Parameters
        ----------
        tautomer: rdkit.Chem.Molecule - A tautomer object

        Returns
        -------
        n_conj: int - the lenght of the conjugate path
        atoms: list - The list of the atoms part of the path

        """

        mf = MolFragmenter(tautomer)
        mf.fragment()
        mf.get_graphs()
        mf.compute_paths()
        path = mf.get_longest_path()
        n_conj = mf.get_conjugate_length(path)
        atoms = mf.get_depiction_atoms(path)
        return n_conj, atoms

    def compute_score(self, mol, returndetails=False, log=False):
        """Return a canonical tautomer by enumerating and scoring all possible tautomers.

        Parameters
        ----------
        mol: :class:`Mol <Chem.rdchem.Mol-class.html>`
            The input molecule

        Returns
        -------
        tau : list :class:`Mol <Chem.rdchem.Mol-class.html>`
            The canonical tautomers
        t_score : list
            The tautomer scores
        """
        t_scores = []
        scores_detail = []
        t_depict = []

        if isinstance(mol, SmallMol):
            mol = mol._mol

        # TODO: Overload the mol parameter to pass a list of pre-enumerated tautomers
        tautomers = self._enumerate_tautomers(mol)

        for t in tautomers:
            tmp_score_details = {'ArRing': 0,
                                 'CarbArRing': 0,
                                 'MatchFeature': [0, []],
                                 'Penalty': [0, []],
                                 'Conjugate': 0}
            smiles = Chem.MolToSmiles(t, isomericSmiles=True)
            if log:
                print('Tautomer: %s', smiles)
            score = 0
            # Add aromatic ring scores
            ssr = Chem.GetSymmSSSR(t)
            for ring in ssr:
                btypes = {t.GetBondBetweenAtoms(*pair).GetBondType() for pair in pairwise(ring)}
                elements = {t.GetAtomWithIdx(idx).GetAtomicNum() for idx in ring}
                if btypes == {BondType.AROMATIC}:
                    if log:
                        print('Score +100 (aromatic ring)')
                    score += 100
                    tmp_score_details['ArRing'] += 1
                    if elements == {6}:
                        if log:
                            print('Score +150 (carbocyclic aromatic ring)')
                        score += 150
                        tmp_score_details['CarbArRing'] += 1

            # Add SMARTS scores, Chem.MolToSmiles(t))
            for tscore in self.scores:

                for match in t.GetSubstructMatches(tscore.smarts):
                    if log:
                        print('Score %+d (%s)' % (tscore.score, tscore.name))
                    score += tscore.score
                    tmp_score_details['MatchFeature'][0] += 1
                    tmp_score_details['MatchFeature'][1].append(tscore.name)

            # Add (P,S,Se,Te)-H scores
            for atom in t.GetAtoms():
                if atom.GetAtomicNum() in {15, 16, 34, 52}:
                    hs = atom.GetTotalNumHs()
                    if hs:
                        if log:
                            print('Score %+d (%s-H bonds)' % (-hs, atom.GetSymbol()))
                        score -= hs
                        tmp_score_details['Penalty'][0] += 1
                        tmp_score_details['Penalty'][1].append(atom.GetSymbol())

            # compute the conjuggate system
            n_conjugate, depictionatoms = self.get_conjugate(t)
            t_depict.append(depictionatoms)

            tmp_score_details['Conjugate'] = n_conjugate
            score += n_conjugate * 2

            scores_detail.append(tmp_score_details)
            t_scores.append(score)
        if returndetails:
            return [SmallMol(tautomer) for tautomer in tautomers], t_scores, t_depict, scores_detail
        return tautomers, t_scores

    @staticmethod
    def filter_tautomers(tautomers, scores, threshold=2):
        """ The function returns the tautomers as rdkit molecule objects based on the scores and the threshold

        Parameters
        ----------
        tautomers: list - List of rdkit.Chem.Molecule of the tautomers identified
        scores: list - List of the scores for each tatutomer
        threshold: int - The threshold value to be used as difference from the highest one

        Returns
        -------
        t_filtered: list - List of rdkit.Chem.Molecule of the tautomers filtered
        """

        tautomers = [mol._mol if isinstance(mol, SmallMol) else mol for mol in tautomers]

        tautomers_sorted = [x for _, x in sorted(zip(scores, tautomers), key=lambda pair: pair[0], reverse=True)]
        scores.sort(reverse=True)
        t_filterd = [t for t, s in zip(tautomers_sorted, scores) if s >= max(scores) - threshold]
        return [SmallMol(rdmol) for rdmol in t_filterd]



import os
import unittest
from tempfile import NamedTemporaryFile
from glob import glob
from moleculekit.home import home
from moleculekit.smallmol.smallmol import SmallMol
from moleculekit.smallmol.smallmollib import SmallMolLib
import rdkit
from pandas import core


SDF_N_MOLS = 100

SDF_IDS_DELETE = [1,10,15,16]
SDF_MOLNAME_DELETE = ['ZINC02583363', 'ZINC86860147', 'ZINC04342657', 'ZINC02023420']
SDF_FIELDS = ['ligname', '_mol']
SDF_LOC_0_99 = 'ZINC02141008'

class _TestSmallMol(unittest.TestCase):

    def setUp(self):
        self.dataDir= home('test-smallmol')

    def test_loaSdffile(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(n_mols, SDF_N_MOLS, 'Molecules not correctly loaded. '
                                             'Expected: {}; Now: {}'.format(SDF_N_MOLS, n_mols))

        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf.gz')
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(n_mols, SDF_N_MOLS, 'Molecules not correctly loaded. '
                                             'Expected: {}; Now: {}'.format(SDF_N_MOLS, n_mols))

    def test_writeSdf(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)

        sdfname = NamedTemporaryFile().name + '.sdf'
        lib.writeSdf(sdfname)

        sdf_exists = os.path.isfile(sdfname)

        self.assertTrue(sdf_exists, msg="The sdf written was not found")

        sdf = SmallMolLib(sdfname)

        self.assertIsInstance(sdf, SmallMolLib, msg="The sdf written was not correctly loaded. Probably the previous"
                                                    "writing went wrong")


    def test_appendSmallMolLib(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        lib2 = SmallMolLib(sdffile)

        lib.appendSmallLib(lib2)

        n_mol2_merged = lib.numMols

        self.assertEqual(n_mol2_merged, SDF_N_MOLS*2, msg="The number of molecules in the SmallMolLib is not as expected."
                                                          "The two sdf were not correctly merged. ")


    def test_appendSmallMol(self):
        mol2file = os.path.join(self.dataDir, 'benzamidine.mol2')
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)
        sm = SmallMol(mol2file)
        lib.appendSmallMol(sm)

        n_mol2_append = lib.numMols

        self.assertEqual(n_mol2_append, SDF_N_MOLS+1, msg="The number of molecules in the SmallMolLib is not as expected."
                                                          "The mol2 were not correctly append. ")

    def test_removeMols(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        mols_ids = SDF_IDS_DELETE
        ref_mols_name = SDF_MOLNAME_DELETE

        mols_name = [ s.ligname for s in lib.getMols(mols_ids) ]

        self.assertListEqual(mols_name, ref_mols_name, msg="The molecules at the given indexes do not match with the"
                                                           "expected")
        lib.removeMols(mols_ids)

        mols_name_now = [ s.ligname for s in lib.getMols(mols_ids) ]

        self.assertFalse(mols_name_now == mols_name, msg="The molecules seem to not be deleted correctly")

    def test_convertToDataFrame(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        df = lib.toDataFrame()

        self.assertIsInstance(df, core.frame.DataFrame, msg="The SmallMolLib object was not correctly converted into pandas"
                                                      "DataFrame")

        cols = df.columns.tolist()
        ref_cols =SDF_FIELDS

        self.assertEqual(cols, ref_cols, msg="The fields in the SmallMolLib object was not the expected one")

        ligname_99 = df.iloc[99][0]
        ref_ligname = SDF_LOC_0_99

        self.assertEqual(ligname_99, ref_ligname, msg="The ligand name found is not the expected one")

    @unittest.skipIf(rdkit.__version__ == '2019.03.4', "new rdkit versions break 2D depiction")
    def test_depict(self):
        import IPython
        refimg = os.path.join(self.dataDir, 'sdf.svg')
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)

        img_name = NamedTemporaryFile().name + '.svg'
        lib.depict(sketch=True, filename=img_name)
        _img = lib.depict(sketch=True, ipython=True)

        refimg_size = os.path.getsize(refimg)
        sm_img_size = os.path.getsize(img_name)

        self.assertIsInstance(_img, IPython.core.display.SVG, msg="The object is not an IPython image as expected")
        self.assertEqual(sm_img_size, refimg_size, msg="The svg image does not have the same size of the reference")


    def test_writeSmiles(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')

        lib = SmallMolLib(sdffile)
        tmpfile = NamedTemporaryFile().name + '.smi'
        lib.writeSmiles(tmpfile)

        with open(tmpfile, 'r') as f:
            filelines = f.readlines()[1:]

        with open(os.path.join(self.dataDir, 'fda_drugs_light.smi'), 'r') as f:
            reflines = f.readlines()[1:]

        self.assertEqual(filelines, reflines)

    def test_readSmiles(self):
        smifile = os.path.join(self.dataDir, 'fda_drugs_light.smi')
        lib = SmallMolLib(smifile)
        assert len(lib) == 100

        smifile = os.path.join(self.dataDir, 'fda_drugs_light.smi.gz')
        lib = SmallMolLib(smifile)
        assert len(lib) == 100



if __name__ == '__main__':
    unittest.main(verbosity=2)


import os
import unittest
from tempfile import NamedTemporaryFile
from glob import glob
from moleculekit.home import home
from moleculekit.smallmol.smallmol import SmallMol
from moleculekit.smallmol.smallmollib import SmallMolLib
from pandas import core
from moleculekit.smallmol.tools.clustering import cluster, getMaximumCommonSubstructure
import numpy as np
import rdkit


CMS_N_ATOMS = 2
CMS_AIDX_MOL0 = [0,1]

# clusters
MACCS_N_CLUSTER = 75
MACCS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 3, 1, 2, 1, 1, 1, 1, 1, 2, 3, 2, 2, 2, 1, 1, 1, 1, 2, 3, 
                            1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 
                            1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

PATHFINGERPRINTS_N_CLUSTER = 82
PATHFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 1, 
                                       1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 
                                       1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

ATOMSFINGERPRINTS_N_CLUSTER = 79
ATOMSFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 1, 1, 
                                        1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 
                                        1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

TORSIONSFINGERPRINTS_N_CLUSTER = 79
TORSIONSFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 
                                           1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 
                                           1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]


CIRCULARFINGERPRINTS_N_CLUSTER = 80
CIRCULARFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 3, 1, 2, 2, 1, 
                                           1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 
                                           1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1]

MCS_N_CLUSTER = 77
MCS_POPULATION_CLUSTER = [2, 1, 2, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 2, 1, 1, 1, 2, 2, 1, 1, 2, 1, 
                          1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 
                          2, 1, 1, 1, 1, 1, 1, 1]

class _TestSmallMol(unittest.TestCase):

    def setUp(self):
        self.dataDir= home('test-smallmol')

    def test_getCommonStructure(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        rd_mols = lib._mols

        cms, cms_idxs, other_idxs = getMaximumCommonSubstructure(rd_mols, returnAtomIdxs=True)

        self.assertIsInstance(cms, rdkit.Chem.rdchem.Mol, msg="The object is not a rdkit Molecule object")

        ref_cms_natoms = CMS_N_ATOMS
        ref_atomidx_mol0 = CMS_AIDX_MOL0
        cms_atomidx_mol0 = cms_idxs[0]
        cms_natoms = len(cms_atomidx_mol0)

        self.assertEqual(cms_atomidx_mol0, ref_atomidx_mol0, msg="The atomidx of the CMS for the mol0 are not the"
                                                                 " expected ones")
        self.assertEqual(cms_natoms, ref_cms_natoms,  msg="The number of atom in the CMS for the mol0 are not the"
                                                                 " expected ones")

    _methods = ['maccs', 'pathFingerprints', 'atomsFingerprints', 'torsionsFingerprints',
                'circularFingerprints', 'shape', 'mcs']
    def test_cluster_maccs(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'maccs', returnDetails=True)

        ref_ncluster = MACCS_N_CLUSTER
        ref_populations = MACCS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()

        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")


    def test_cluster_pathFingerprints(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'pathFingerprints', returnDetails=True)

        ref_ncluster = PATHFINGERPRINTS_N_CLUSTER
        ref_populations = PATHFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()

        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")


    def test_cluster_atomsFingerprints(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'atomsFingerprints', returnDetails=True)

        ref_ncluster = ATOMSFINGERPRINTS_N_CLUSTER
        ref_populations = ATOMSFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()

        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")

    def test_cluster_torsionsFingerprints(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'torsionsFingerprints', returnDetails=True)

        ref_ncluster = TORSIONSFINGERPRINTS_N_CLUSTER
        ref_populations = TORSIONSFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()
        
        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")

    def test_cluster_circularFingerprints(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'circularFingerprints', returnDetails=True)

        ref_ncluster = CIRCULARFINGERPRINTS_N_CLUSTER
        ref_populations = CIRCULARFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()

        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")

    def test_cluster_shape(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'shape', returnDetails=True)

        n_clusters = det['numClusters']

        self.assertIsInstance(n_clusters, np.int64, msg="None valid number of clusters")

    def test_cluster_mcs(self):
        sdffile = os.path.join(self.dataDir, 'fda_drugs_light.sdf')
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        cl, det = cluster(mols, 'mcs', returnDetails=True)

        ref_ncluster = MCS_N_CLUSTER
        ref_populations = MCS_POPULATION_CLUSTER

        ncluster = det['numClusters']
        population = det['populations'].tolist()

        self.assertEqual(ncluster, ref_ncluster, msg="The number of cluster identified are not as expected")
        self.assertEqual(population, ref_populations, msg="The population fo the cluster are not the expected one")

    def test_restrained_embedding(self):
        from moleculekit.smallmol.tools.restrainedembed import restrainedEmbed
        import numpy as np

        atomPos = np.load(os.path.join(self.dataDir, 'restrain_positions.npy'), allow_pickle=True).item()
        refcoords = np.load(os.path.join(self.dataDir, 'restrain_final_coords.npy'), allow_pickle=True)
        sml = SmallMolLib(os.path.join(self.dataDir, 'restrain_molecule.sdf'))
        sm = sml[0]
        
        restrainedEmbed(sm._mol, atomPos)

        assert np.allclose(refcoords, sm._coords)

    # TODO: Add here tests for tautomers


if __name__ == '__main__':
    unittest.main(verbosity=2)


import numpy as np
import logging
logger = logging.getLogger(__name__)


def sequenceStructureAlignment(mol, ref, molseg=None, refseg=None, maxalignments=10, nalignfragment=1):
    """ Aligns two structures by their longests sequences alignment

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule we want to align
    ref : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we want to align
    molseg : str
        The segment of `mol` we want to align
    refseg : str
        The segment of `ref` we want to align to
    maxalignments : int
        The maximum number of alignments we want to produce
    nalignfragment : int
        The number of fragments used for the alignment.

    Returns
    -------
    mols : list
        A list of Molecules each containing a different alignment.
    """
    from moleculekit.util import ensurelist
    try:
        from Bio import pairwise2
    except ImportError as e:
        raise ImportError('You need to install the biopython package to use this function. Try using `conda install biopython`.')
    from Bio.SubsMat import MatrixInfo as matlist

    if len([x for x in np.unique(mol.altloc) if len(x)]) > 1:
        raise RuntimeError('Alternative atom locations detected in `mol`. Please remove these before calling this function.')
    if len([x for x in np.unique(ref.altloc) if len(x)]) > 1:
        raise RuntimeError('Alternative atom locations detected in `ref`. Please remove these before calling this function.')

    seqmol = mol.sequence()
    seqref = ref.sequence()

    if molseg is None:
        msg = 'No segment was specified by the user for `mol`'
        if len(seqmol) > 1:
            msg += ' and multiple segments ({}) were detected. '.format(list(seqmol.keys()))
        else:
            msg += '. '
        msg += 'Alignment will be done on all protein segments.'
        logger.info(msg)
        seqmol = mol.sequence(noseg=True)
        molseg = list(seqmol.keys())[0]
    if refseg is None:
        msg = 'No segment was specified by the user for `ref`'
        if len(seqref) > 1:
            msg += ' and multiple segments ({}) were detected. '.format(list(seqref.keys()))
        else:
            msg += '. '
        msg += 'Alignment will be done on all protein segments.'
        logger.info(msg)
        seqref = ref.sequence(noseg=True)
        refseg = list(seqref.keys())[0]

    def getSegIdx(m, mseg):
        # Calculate the atoms which belong to the selected segments
        if isinstance(mseg, str) and mseg == 'protein':
            msegidx = m.atomselect('protein and name CA')
        else:
            msegidx = np.zeros(m.numAtoms, dtype=bool)
            for seg in ensurelist(mseg):
                msegidx |= (m.segid == seg) & (m.name == 'CA')
        return np.where(msegidx)[0]
    molsegidx = getSegIdx(mol, molseg)
    refsegidx = getSegIdx(ref, refseg)

    # Create fake residue numbers for the selected segment
    from moleculekit.util import sequenceID
    molfakeresid = sequenceID((mol.resid[molsegidx], mol.insertion[molsegidx], mol.chain[molsegidx]))
    reffakeresid = sequenceID((ref.resid[refsegidx], ref.insertion[refsegidx], ref.chain[refsegidx]))

    # TODO: Use BLOSUM62?
    alignments = pairwise2.align.globaldx(seqref[refseg], seqmol[molseg], matlist.blosum62)
    numaln = len(alignments)

    if numaln > maxalignments:
        logger.warning('{} alignments found. Limiting to {} as specified in the `maxalignments` argument.'.format(numaln, maxalignments))

    alignedstructs = []
    for i in range(min(maxalignments, numaln)):
        refaln = np.array(list(alignments[i][0]))
        molaln = np.array(list(alignments[i][1]))
 
        # By doing cumsum we calculate how many letters were before the current letter (i.e. residues before current)
        residref = np.cumsum(refaln != '-') - 1  # Start them from 0
        residmol = np.cumsum(molaln != '-') - 1  # Start them from 0

        # Find the region of maximum alignment between the molecules
        dsig = np.hstack(([False], (refaln != '-') & (molaln != '-'), [False])).astype(int)
        dsigdiff = np.diff(dsig)
        startIndex = np.where(dsigdiff > 0)[0]
        endIndex = np.where(dsigdiff < 0)[0]
        duration = endIndex - startIndex
        duration_sorted = np.sort(duration)[::-1]

        _list_starts = []
        _list_finish = []
        for n in range(nalignfragment):
            if n == len(duration):
                break
            idx = np.where(duration == duration_sorted[n])[0]
            start = startIndex[idx][0]
            finish = endIndex[idx][0]
            _list_starts.append(start)
            _list_finish.append(finish)

        # Get the "resids" of the aligned residues only
        refalnresid = np.concatenate([ residref[start:finish] for start, finish in zip(_list_starts,_list_finish)])
        molalnresid = np.concatenate([ residmol[start:finish] for start, finish in zip(_list_starts, _list_finish) ])
        refidx = []
        for r in refalnresid:
            refidx += list(refsegidx[reffakeresid == r])
        molidx = []
        for r in molalnresid:
            molidx += list(molsegidx[molfakeresid == r])        

        molboolidx = np.zeros(mol.numAtoms, dtype=bool)
        molboolidx[molidx] = True
        refboolidx = np.zeros(ref.numAtoms, dtype=bool)
        refboolidx[refidx] = True

        start_residues = np.concatenate([ mol.resid[molsegidx[molfakeresid == residmol[r]]] for r in _list_starts])
        finish_residues = np.concatenate([ mol.resid[molsegidx[molfakeresid == residmol[r-1]]] for r in _list_finish])
        logger.info('Alignment #{} was done on {} residues: mol segid {} resid {}'.format(
            i, len(refalnresid), np.unique(mol.segid[molidx])[0], ', '.join(['{}-{}'.format(s,f) for s, f in zip(start_residues,finish_residues)])  ))

        alignedmol = mol.copy()
        alignedmol.align(molboolidx, ref, refboolidx)
        alignedstructs.append(alignedmol)

    return alignedstructs


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricCoordinate(Projection):
    """ Creates a MetricCoordinate object that calculates the atom coordinates from a set of trajectories.

    Parameters
    ----------
    atomsel : str
        Atom selection string for the atoms whose coordinates we want to calculate.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    refmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we will align.
    trajalnsel : str, optional
        Atom selection string for the trajectories from which to align to the reference structure.
        If it's None and a `refmol` is passed it will default to 'protein and name CA'.
    refalnsel : str, optional
        Atom selection string for `refmol` from which to align to the reference structure. If None, it defaults to the
        same as `trajalnsel`.
    centersel : str, optional
        Atom selection string around which to wrap the simulation.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    pbc : bool
        Enable or disable coordinate wrapping based on periodic boundary conditions.

    Returns
    -------
    metr : MetricCoordinate object
    """
    def __init__(self, atomsel, refmol=None, trajalnsel=None, refalnsel=None, centersel='protein', pbc=True):
        super().__init__()

        if atomsel is None:
            raise ValueError('Atom selection cannot be None')
        
        self._trajalnsel = trajalnsel
        self._centersel = centersel
        self._atomsel = atomsel
        self._pbc = pbc

        self._refmol = refmol
        self._refalnsel = refalnsel
        if self._refmol is not None:
            if self._trajalnsel is None:
                self._trajalnsel = 'protein and name CA'
            self._refmol = refmol.copy()
            self._cache['refalnsel'] = self._refmol.atomselect(self._refalnsel if self._refalnsel is not None else self._trajalnsel)

    def _calculateMolProp(self, mol, props='all'):
        props = ('trajalnsel', 'atomsel', 'centersel') if props == 'all' else props
        res = {}
        if 'trajalnsel' in props:
            res['trajalnsel'] = None
            if self._trajalnsel is not None:
                res['trajalnsel'] = mol.atomselect(self._trajalnsel)
                if np.sum(res['trajalnsel']) == 0:
                    raise RuntimeError('Alignment selection resulted in 0 atoms.')                
        if 'atomsel' in props:
            res['atomsel'] = mol.atomselect(self._atomsel)
            if np.sum(res['atomsel']) == 0:
                raise RuntimeError('Atom selection resulted in 0 atoms.')
        if 'centersel' in props:
            res['centersel'] = mol.atomselect(self._centersel)
            if np.sum(res['centersel']) == 0:
                raise RuntimeError('Centering selection resulted in 0 atoms.')
        return res

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        mol = mol.copy()
        if self._pbc:
            mol.wrap(getMolProp('centersel'))

        trajalnsel = getMolProp('trajalnsel')

        if trajalnsel is not None:
            if self._refmol is None:
                mol.align(trajalnsel)
            else:
                mol.align(trajalnsel, refmol=self._refmol, refsel=getMolProp('refalnsel'))

        coords = np.transpose(mol.coords[getMolProp('atomsel')], axes=(2, 1, 0)).reshape(mol.numFrames, -1)
        return coords

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        atomidx = np.where(getMolProp('atomsel'))[0]
        from pandas import DataFrame
        types = []
        indexes = []
        description = []
        for xyz in ('X', 'Y', 'Z'):
            for i in atomidx:
                types += ['coordinate']
                indexes += [i]
                description += ['{} coordinate of {} {} {}'.format(xyz, mol.resname[i], mol.resid[i], mol.name[i])]
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


import unittest
class _TestMetricCoordinate(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))

        self.mol = mol

    def test_project(self):
        ref = self.mol.copy()
        ref.coords = np.atleast_3d(ref.coords[:, :, 0])
        metr = MetricCoordinate('protein and name CA')
        data = metr.project(self.mol)

        lastcoors = np.array([-24.77000237, -27.76000023, -30.44000244, -33.65000153,
                              -33.40999985, -36.32000351, -36.02000427, -36.38000107,
                              -39.61000061, -41.01000214, -43.80000305, -45.56000137,
                              -45.36000061, -47.13000488, -49.54000473, -50.6000061 ,
                              -50.11000061, -52.15999985, -55.1400032 , -55.73000336], dtype=np.float32)
        assert np.all(np.abs(data[-1, -20:] - lastcoors) < 0.001), 'Coordinates calculation is broken'

    def test_project_align(self):
        ref = self.mol.copy()
        ref.coords = np.atleast_3d(ref.coords[:, :, 0])
        metr = MetricCoordinate('protein and name CA', ref)
        data = metr.project(self.mol)

        lastcoors = np.array([6.79283285, 5.55226946, 4.49387407, 2.94484425,
                              5.36937141, 3.18590879, 5.75874281, 5.48864174,
                              1.69625032, 1.58790839, 0.57877392, -2.66498065,
                              -3.70919156, -3.33702421, -5.38465405, -8.43286991,
                              -8.15859032, -7.85062265, -10.92551327, -13.70733166], dtype=np.float32)
        assert np.all(np.abs(data[-1, -20:] - lastcoors) < 0.001), 'Coordinates calculation is broken'


if __name__ == '__main__':
    unittest.main(verbosity=2)




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import math
import numpy as np
from rdkit import Chem
from moleculekit.smallmol.smallmol import SmallMol
import gzip
import logging


logger = logging.getLogger(__name__)


def smiReader(file, removeHs, fixHs, isgzip=False):
    from tqdm import tqdm
    if isgzip:
        with gzip.open(file, 'rb') as f:
            return smiReader(f, removeHs=removeHs, fixHs=fixHs, isgzip=False)

    if isinstance(file, str):
        with open(file, 'r') as f:
            return smiReader(f, removeHs, fixHs)

    decode = False
    if isinstance(file, gzip.GzipFile):
        decode = True
            
    lines = file.readlines()[1:]
    mols = []
    for i, line in enumerate(tqdm(lines)):
        if decode:
            line = line.decode('utf-8')
        smi, name = line.strip().split()
        try:
            mols.append(SmallMol(smi, removeHs=removeHs, fixHs=fixHs))
        except Exception as e:
            logger.warning('Failed to load molecule with name {} with error {}. Skipping to next molecule.'.format(name, e))
    return mols

def sdfReader(file, removeHs, fixHs, sanitize, isgzip=False):
    from tqdm import tqdm
    from moleculekit.util import tempname

    if isgzip:
        with gzip.open(file, 'rb') as f:
            # SDMolSupplier does not support file handles, need to write temp file
            file = tempname(suffix='.sdf')
            with open(file, 'wb') as fout:
                fout.write(f.read())

    supplier = Chem.SDMolSupplier(file, removeHs=removeHs, sanitize=sanitize)
    mols = []
    countfailed = 0
    for mol in tqdm(supplier):
        if mol is None:
            countfailed += 1
            continue
        try:
            mols.append(SmallMol(mol, removeHs=removeHs, fixHs=fixHs))
        except:
            if mol.HasProp('_Name'):
                name = mol.GetProp('_Name')
            countfailed += 1
            logger.warning('Failed to load molecule{}. Skipping to next molecule.'.format(' with name {}'.format(name)))
    if countfailed:
        logger.info('Failed to load {}/{} molecules'.format(countfailed, len(supplier)))
    return mols

class SmallMolLib(object):
    """
    Class to manage ligands databases (sdf). Ligands are stored as moleculekit.smallmol.smallmol.SmallMol objects and
    fields type in the sdf are stored in a list

    Parameters
    ----------
    lib_file: str
        The sdf or smi file path
    removeHs: bool
        If True, the hydrogens of the molecules will be removed
    fixHs: bool
        If True, the hydrogens are added and optimized

    Example
    -------
    >>> import os
    >>> from moleculekit.home import home
    >>> lib = SmallMolLib(os.path.join(home(dataDir='test-smallmol'), 'fda_drugs_light.sdf'))
    >>> lib.numMols
    100

    .. rubric:: Methods
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMolLib
       :methods:
    .. rubric:: Attributes
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMolLib
       :attributes:

    Attributes
    ----------
    numMols: int
        Number of SmallMol molecules

    """

    def __init__(self, libfile=None, removeHs=False, fixHs=True, sanitize=True):  # , n_jobs=1
        if libfile is not None:
            self._mols = self._loadLibrary(libfile, removeHs=removeHs, fixHs=fixHs, sanitize=sanitize, ext=None)


    def _loadLibrary(self, libfile, removeHs=False, fixHs=True, sanitize=True, ext=None):
        isgzip = False
        if ext == None:
            ext = os.path.splitext(libfile)[-1] 
        if ext == '.gz':
            isgzip = True
            ext = os.path.splitext(os.path.splitext(libfile)[-2])[-1]

        if ext == '.sdf':
            return sdfReader(libfile, removeHs, fixHs, sanitize, isgzip)
        elif ext == '.smi':
            return smiReader(libfile, removeHs, fixHs, isgzip)
        else:
            raise RuntimeError('Invalid file extension {}. Could not read it.'.format(ext))

    @property
    def numMols(self):
        """
        Returns the number of molecules
        """
        return len(self._mols)

    def getMols(self, ids=None):
        """
        Returns the SmallMol objects that corresponds ot the indexes of the list passed

        Parameters
        ----------
        ids: list
            The index list of the molecules to return

        Returns
        -------
        smallmollist: list
            The list of SmallMol objects

        Example
        -------
        >>> lib2 = lib.getMols([1,2,3])
        >>> len(lib2)
        3
        """

        if ids is None:
            return self._mols
        if not isinstance(ids, list):
            raise TypeError("The argument ids {} should be list".format(type(ids)))

        return np.array(self._mols)[ids]

    def writeSdf(self, sdf_name, fields=None):
        """
        Writes an sdf file with molecules stored. Is it possible also to manage which field will be written

        Parameters
        ----------
        sdf_name: str
            The ouput sdf filename
        fields: list
            A list of the fields to write. If None all are saved
        """

        from rdkit.Chem import SDWriter

        writer = SDWriter(sdf_name)
        if fields is not None:
            if not isinstance(fields, list):
                raise TypeError("The fields argument {} should be a list".format(type(fields)))
            writer.SetProps(fields)

        for m in self._mols:
            writer.write(m._mol)

    def writeSmiles(self, smi_name, explicitHs=True, names=False, header=None):
        """
        Writes a smi file with molecules stored. Is it possible to specify the header of the smi file. The name of the
        ligands can be their ligand name or a sequential ID.

        Parameters
        ----------
        smi_name: str
            The ouput smi filename
        names: bool
            Set as True to use the own ligand name for each ligand. Otherwise a sequential ID will be used
        header: str
            The header of the smi file. If is None the smi filename will be used.
        """

        smi_name = os.path.splitext(smi_name)[0] + '.smi'

        f = open(smi_name, 'w')

        if header is None:
            header = os.path.splitext(smi_name)[0]

        f.write(header + '\n')

        for n, sm in enumerate(self.getMols()):
            smi = sm.toSMILES(explicitHs=explicitHs)
            name = n if not names else sm.ligname
            f.write(smi + ' {} \n'.format(name))

        f.close()

    def appendSmallLib(self, smallLib, strictField=False, strictDirection=1):
        """
        Merge two moleculekit.smallmol.smallmol.SmallMolLib objects

        Parameters
        ----------
        smallLib: moleculekit.smallmol.smallmol.SmallMolLib
            The new SmallMolLib to merge
        """
        self._mols += smallLib._mols


    def appendSmallMol(self, smallmolecule, strictField=False, strictDirection=0):
        """
        Adds a new moleculekit.smallmol.smallmol.SmallMol object in the current SmallMolLib object

        Parameters
        ---------
        smallmol: moleculekit.smallmol.smallmol.SmallMol
            The SmallMol object to add
        """
        self._mols = np.append(self._mols, smallmolecule)

    def removeMols(self, ids):
        """
        Removes the moleculekit.smallmol.smallmol.SmallMol object based on the indexes in the list

        Parameters
        ----------
        ids: list
            The list of molecules index to remove from the SmallMolLib
        """

        if not isinstance(ids, list):
            raise TypeError('The argument ids {} is not valid. Should be list'.format(type(ids)))
        _oldNumMols = self.numMols
        self._mols = np.delete(self._mols, ids)

        logger.warning("[num mols before deleting: {}]. The molecules {} were removed, now the number of "
                       "molecules are {} ".format(_oldNumMols, ids, self.numMols))

    def toDataFrame(self, fields=None, molAsImage=True, sketch=True):
        """
        Returns a pandas.DataFrame of the SmallMolLib object.

        Parameters
        ----------
        fields: list
            The list of fields to convert into a pandas DataFrame column
        molAsImage: bool
            If True, the rdkit.Chem.rdchem.Mol is converted into an image
        sketch: bool
            If True, the molecule are rendered to be 2D

        Returns
        -------
        dataframe: pandas.DataFrame
            The pandas DataFrame
        """
        from rdkit.Chem import PandasTools
        from rdkit.Chem.AllChem import Compute2DCoords
        import pandas as pd
        from copy import deepcopy

        if fields is not None:
            if not isinstance(fields, list):
                raise TypeError('The argument fields passed {} should be a list '.format(type(fields)))
        else:
            fields = ['ligname', '_mol'] if molAsImage else ['ligname']

        records = []
        indexes = []
        for i, m in enumerate(self._mols):
            row = dict((f, m.__getattribute__(f)) for f in fields)
            if sketch:
                mm = deepcopy(m._mol)
                Compute2DCoords(mm)
                row['_mol'] = mm
            records.append(row)
            indexes.append(i)

        df = pd.DataFrame(records, columns=fields, index=indexes)
        if molAsImage:
            Chem.PandasTools.ChangeMoleculeRendering(df)
        return df

    def copy(self):
        """
        Returns a copy of the SmallMolLib object
        """
        from copy import deepcopy
        return deepcopy(self)

    def __len__(self):
        return len(self._mols)

    def __getitem__(self, item):
        return self._mols[item]

    def __iter__(self):
        _mols = self._mols
        for smallmol in _mols:
            yield smallmol

    def __str__(self):
        _mols = self._mols

        return ('Stack of Small molecules.'
                '\n\tContains {} Molecules.'
                '\n\tSource file: "{}".').format(len(_mols), self._sdffile)

    def depict(self, ids=None, sketch=True, filename=None, ipython=False, optimize=False, optimizemode='std',
               removeHs=True,  legends=None, highlightAtoms=None, mols_perrow=3):

        """
        Depicts the molecules into a grid. It is possible to save it into an svg file and also generates a
        jupiter-notebook rendering

        Parameters
        ----------
        ids: list
            The index of the molecules to depict
        sketch: bool
            Set to True for 2D depiction
        filename: str
            Set the filename for the svg file
        ipython: bool
            Set to True to return the jupiter-notebook rendering
        optimize: bool
            Set to True to optimize the conformation. Works only with 3D.
        optimizemode: ['std', 'mmff']
            Set the optimization mode for 3D conformation
        removeHs: bool
            Set to True to hide hydrogens in the depiction
        legends: str
            The name to used for each molecule. Can be 'names':the name of themselves; or 'items': a incremental id
        highlightAtoms: list
            A List of atom to highligh for each molecule. It can be also a list of atom list, in this case different
            colors will be used
        mols_perrow: int
            The number of molecules to depict per row of the grid

        Returns
        -------
            ipython_svg: SVG object if ipython is set to True

        """
        from rdkit.Chem.AllChem import Compute2DCoords, EmbedMolecule, MMFFOptimizeMolecule, ETKDG
        from rdkit.Chem import RemoveHs
        from moleculekit.smallmol.util import depictMultipleMols

        if sketch and optimize:
            raise ValueError('Impossible to use optmization in  2D sketch representation')

        if legends is not None and legends not in ['names', 'items']:
            raise ValueError('The "legends" should be "names" or "items"')

        _smallmols = self.getMols(ids)

        if ids is None:
            _mols = [m._mol for m in self._mols]
        else:
            _mols = [m._mol for m in self.getMols(ids)]

        if highlightAtoms is not None:
            if len(highlightAtoms) != len(_mols):
                raise ValueError('The highlightAtoms {} should have the same length of the '
                                 'mols {}'.format(len(highlightAtoms), len(_mols)))

        if sketch:
            for _m in _mols:
                Compute2DCoords(_m)

        if removeHs:
            _mols = [RemoveHs(_m) for _m in _mols]

        # activate 3D coords optimization
        if optimize:
            if optimizemode == 'std':
                for _m in _mols:
                    EmbedMolecule(_m)
            elif optimizemode == 'mmff':
                for _m in _mols:
                    MMFFOptimizeMolecule(_m, ETKDG())

        legends_list = []
        if legends == 'names':
            legends_list = [_m.getProp('ligname') for _m in _smallmols]
        elif legends == 'items':
            legends_list = [str(n+1) for n in range(len(_smallmols))]

        return depictMultipleMols(_mols, ipython=ipython, legends=legends_list, highlightAtoms=highlightAtoms,
                                  filename=filename, mols_perrow=mols_perrow)

SmallMolStack = SmallMolLib

if __name__ == '__main__':
    import doctest
    import os
    from moleculekit.home import home

    lib = SmallMolLib(os.path.join(home(dataDir='test-smallmol'), 'fda_drugs_light.sdf'))
    doctest.testmod(extraglobs={'lib': lib})


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging
logger = logging.getLogger(__name__)


class MetricSasa(Projection):
    """ Calculate solvent accessible surface area of a molecule.

    Implementation and documentation taken from MDtraj shrake_rupley code.

    Parameters
    ----------
    sel : str
        Atom selection string for atoms or residues for which to calculate the SASA.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    filtersel : str
        Keep only the selected atoms in the system. All other atoms will be removed and will
        not contribute to the SASA calculation. Keep in mind that the SASA of an atom or residue
        is affected by the presence of other atoms around it so this will change the SASA of the remaining atoms.
    probeRadius : float
        The radius of the probe, in Angstrom.
    numSpherePoints : int
        The number of points representing the surface of each atom, higher values lead to more accuracy.
    mode : str
        In mode == 'atom', the extracted areas are resolved per-atom. In mode == 'residue', this is consolidated down
        to the per-residue SASA by summing over the atoms in each residue.

    Returns
    -------
    metr : MetricSasa object
    """
    def __init__(self, sel='protein', filtersel='all', probeRadius=1.4, numSpherePoints=960, mode='atom'):
        super().__init__()

        self._probeRadius = probeRadius
        self._numSpherePoints = numSpherePoints
        self._mode = mode
        self._sel = sel
        self._filtersel = filtersel

    def _calculateMolProp(self, mol, props='all'):
        props = ('radii', 'atom_mapping', 'sel', 'filtersel', 'tokeep') if props == 'all' else props
        res = {}

        sel = mol.atomselect(self._sel)
        selidx = np.where(sel)[0]
        if 'sel' in props:
            res['sel'] = sel
        
        filtersel = mol.atomselect(self._filtersel)
        filterselidx = np.where(filtersel)[0]
        if 'filtersel' in props:
            res['filtersel'] = filtersel

        if len(np.setdiff1d(selidx, filterselidx)) != 0:
            raise RuntimeError('Some atoms selected by `sel` are not selected by `filtersel` and thus would not be calculated. Make sure `sel` is a subset of `filtersel`.')

        if 'tokeep' in props:
            filterselmod = filtersel.copy().astype(int)
            filterselmod[filterselmod == 0] = -1
            filterselmod[filtersel] = np.arange(np.count_nonzero(filtersel))
            res['tokeep'] = filterselmod[sel]

        if 'radii' in props:
            _ATOMIC_RADII = {'C': 1.5, 'F': 1.2, 'H': 0.4, 'N': 1.10, 'O': 1.05, 'S': 1.6, 'P': 1.6}
            elements = [n[0] for n in mol.name[filtersel]]
            atom_radii = np.vectorize(_ATOMIC_RADII.__getitem__)(elements)
            res['radii'] = np.array(atom_radii, np.float32) + self._probeRadius

        if 'atom_mapping' in props:
            if self._mode == 'atom':
                res['atom_mapping'] = np.arange(np.sum(filtersel), dtype=np.int32)
            elif self._mode == 'residue':
                from moleculekit.util import sequenceID
                res['atom_mapping'] = sequenceID((mol.resid[filtersel], mol.chain[filtersel], mol.segid[filtersel])).astype(np.int32)
            else:
                raise ValueError('mode must be one of "residue", "atom". "{}" supplied'.format(self._mode))

        return res

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        radii = getMolProp('radii')
        atom_mapping = getMolProp('atom_mapping')
        sel = getMolProp('sel')
        filtersel = getMolProp('filtersel')
        tokeep = getMolProp('tokeep')
        tokeep = np.unique(atom_mapping[tokeep])

        xyz = np.swapaxes(np.swapaxes(np.atleast_3d(mol.coords[filtersel, :, :]), 1, 2), 0, 1)
        xyz = np.array(xyz.copy(), dtype=np.float32) / 10  # converting to nm

        try:
            from mdtraj.geometry._geometry import _sasa as sasa
        except ImportError:
            raise ImportError('To calculate SASA you need to install mdtraj with `conda install mdtraj -c omnia`')

        out = np.zeros((mol.numFrames, atom_mapping.max() + 1), dtype=np.float32)
        sasa(xyz, radii / 10, int(self._numSpherePoints), atom_mapping, out)  # Divide radii by 10 for nm
        return out[:, tokeep]

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        atom_mapping = getMolProp('atom_mapping')
        tokeep = getMolProp('tokeep')
        atomsel = getMolProp('sel')

        if self._mode == 'atom':
            atomidx = np.where(atomsel)[0]
        elif self._mode == 'residue':
            atom_mapping = atom_mapping[tokeep]
            _, firstidx = np.unique(atom_mapping, return_index=True)
            atomidx = np.where(atomsel)[0][firstidx]
        else:
            raise ValueError('mode must be one of "residue", "atom". "{}" supplied'.format(self._mode))

        from pandas import DataFrame
        types = []
        indexes = []
        description = []
        for i in atomidx:
            types += ['SASA']
            indexes += [i]
            description += ['SASA of {} {} {}'.format(mol.resname[i], mol.resid[i], mol.name[i])]
        return DataFrame({'type': types, 'atomIndexes': indexes, 'description': description})


import unittest
class _TestMetricSasa(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path
        
        mol = Molecule(path.join(home(dataDir='test-projections'), 'trajectory', 'filtered.pdb'))
        mol.read(path.join(home(dataDir='test-projections'), 'trajectory', 'traj.xtc'))
        mol.dropFrames(keep=[0, 1])  # Keep only two frames because it's super slow
        self.mol = mol

    def test_sasa_atom(self):
        from os import path
        from moleculekit.home import home

        metr = MetricSasa(mode='atom')
        sasaA = metr.project(self.mol.copy())
        sasaA_ref = np.load(path.join(home(dataDir='test-projections'), 'metricsasa', 'sasa_atom.npy'))
        assert np.allclose(sasaA, sasaA_ref, atol=7e-4)

    def test_sasa_residue(self):
        from os import path
        from moleculekit.home import home

        metr = MetricSasa(mode='residue')
        sasaR = metr.project(self.mol.copy())
        sasaR_ref = np.load(path.join(home(dataDir='test-projections'), 'metricsasa', 'sasa_residue.npy'))
        assert np.allclose(sasaR, sasaR_ref, atol=3e-3)

    def test_set_diff_error(self):
        try:
            metr = MetricSasa(mode='atom', sel='index 3000', filtersel='not index 3000')
            metr._calculateMolProp(self.mol.copy())
        except RuntimeError:
            print('Correctly threw a runtime error for bad selections')
        else:
            raise AssertionError('This should throw an error as the selected atom does not exist in the filtered system')
     
    def test_selection_and_filtering(self):
        from os import path
        from moleculekit.home import home

        sasaR_ref = np.load(path.join(home(dataDir='test-projections'), 'metricsasa', 'sasa_atom.npy'))

        metr = MetricSasa(mode='atom', sel='index 3000') # Get just the SASA of the 3000th atom
        sasaR = metr.project(self.mol.copy())
        assert np.allclose(sasaR, sasaR_ref[:, [3000]], atol=3e-3), 'SASA atom selection failed to give same results as without selection'

        metr = MetricSasa(mode='atom', sel='index 3000', filtersel='index 3000') # Get just the SASA of the 3000th atom, remove all else
        sasaR = metr.project(self.mol.copy())
        assert not np.allclose(sasaR, sasaR_ref[:, [3000]], atol=3e-3), 'SASA filtering gave same results as without filtering. Bad.'

    def test_mappings(self):
        metr = MetricSasa(mode='atom')
        mapping = metr.getMapping(self.mol)
        assert np.array_equal(mapping.atomIndexes, np.arange(4480))

        metr = MetricSasa(mode='residue')
        mapping = metr.getMapping(self.mol)
        ref = np.array([0,   17,   27,   39,   49,   56,   75,   99,  113,  132,  152, 167,  189,  211,  222,  241,  256,  268,  290,  304,  319,  343,
                        358,  377,  396,  411,  422,  443,  463,  484,  500,  515,  522, 533,  545,  555,  570,  589,  596,  613,  624,  638,  662,  679,
                        695,  712,  731,  751,  775,  797,  808,  822,  839,  854,  873, 892,  902,  909,  919,  930,  949,  968,  979,  991, 1015, 1039,
                        1055, 1074, 1088, 1098, 1108, 1125, 1135, 1154, 1173, 1194, 1208, 1222, 1246, 1258, 1280, 1294, 1314, 1328, 1343, 1357, 1369, 1388,
                        1407, 1423, 1447, 1466, 1473, 1495, 1512, 1523, 1547, 1561, 1585, 1606, 1621, 1645, 1659, 1678, 1693, 1715, 1734, 1745, 1762, 1781,
                        1796, 1818, 1837, 1858, 1877, 1894, 1908, 1932, 1953, 1967, 1991, 2015, 2030, 2044, 2063, 2075, 2099, 2111, 2130, 2140, 2159, 2176,
                        2198, 2217, 2239, 2261, 2275, 2291, 2301, 2321, 2332, 2344, 2365, 2384, 2401, 2415, 2431, 2441, 2460, 2474, 2486, 2510, 2525, 2539,
                        2549, 2559, 2570, 2589, 2608, 2625, 2635, 2642, 2663, 2685, 2692, 2716, 2732, 2746, 2753, 2777, 2784, 2798, 2817, 2839, 2854, 2861,
                        2878, 2892, 2903, 2919, 2938, 2955, 2971, 2987, 3001, 3020, 3034, 3053, 3069, 3084, 3108, 3122, 3138, 3148, 3170, 3182, 3193, 3207,
                        3231, 3250, 3274, 3293, 3307, 3319, 3333, 3350, 3370, 3380, 3390, 3397, 3418, 3440, 3454, 3466, 3481, 3488, 3510, 3534, 3541, 3553,
                        3563, 3573, 3588, 3595, 3607, 3618, 3625, 3632, 3646, 3666, 3682, 3699, 3721, 3732, 3746, 3766, 3780, 3794, 3818, 3842, 3863, 3880,
                        3897, 3904, 3923, 3939, 3950, 3974, 3981, 3996, 4003, 4013, 4025, 4049, 4061, 4068, 4090, 4111, 4118, 4138, 4159, 4173, 4190, 4206,
                        4226, 4250, 4269, 4291, 4313, 4337, 4356, 4373, 4395, 4411, 4430, 4442, 4459])
        assert np.array_equal(mapping.atomIndexes,ref)



if __name__ == '__main__':
    unittest.main(verbosity=2)



import os
import unittest
from tempfile import NamedTemporaryFile
from glob import glob
from moleculekit.home import home
from moleculekit.molecule import Molecule
from moleculekit.smallmol.smallmol import SmallMol
import rdkit
from rdkit.Chem import MolFromSmiles

BENZAMIDINE_N_ATOMS = 18
BENZAMIDINE_N_HEAVYATOMS = 9
BENZAMIDINE_BONDTYPES = ['ar', 'ar', '1', 'ar', '1', 'ar', '1', 'ar', '1', 'ar', '1', '1', '2', '1', '1', '1', '1', '1']
BENZAMIDINE_BOND_ATOMS = [[0, 1], [0, 5], [0, 6], [1, 2], [1, 7], [2, 3], [2, 8], [3, 4], [3, 9], [4, 5], [4, 10],
                          [5, 11], [6, 12], [6, 13], [12, 16], [12, 17], [13, 14], [13, 15]]

LIGAND_N_ATOMS = 64
LIGAND_N_HEAVYATOMS = 35

SMILE_SMI = 'c1ccccc1O'
SMILE_N_ATOMS = 13

SDF_N_MOLS = 100

PHENOL_ELEMENT_IDX_1 = 'C'
PHENOL_ELEMENT_NEIGHBORS_OX = [5, 12]
PHENOL_BTYPES_OX = [1, 1]

CHIRAL_SMI = 'C[C@H](Cl)F'
CHIRAL_DETAILS = [('C1', 'S')]

FOUNDBOND_SMI = 'C=CN'

class _TestSmallMol(unittest.TestCase):

    def setUp(self):
        self.dataDir= home('test-smallmol')
        self.benzamidine_mol2 = os.path.join(self.dataDir, 'benzamidine.mol2')

    def test_loadMol2file(self):
        sm = SmallMol(self.benzamidine_mol2)
        n_atoms = sm.numAtoms
        self.assertEqual(n_atoms, BENZAMIDINE_N_ATOMS, 'Atoms not correctly loaded. '
                                                        'Expected: {}; Now: {}'.format(BENZAMIDINE_N_ATOMS, n_atoms))

    def test_loadPdbfile(self):
        pdbfile = os.path.join(self.dataDir, 'ligand.pdb')
        sm = SmallMol(pdbfile)
        n_atoms = sm.numAtoms
        self.assertEqual(n_atoms, LIGAND_N_ATOMS, 'Atoms not correctly loaded. '
                                                        'Expected: {}; Now: {}'.format(LIGAND_N_ATOMS, n_atoms))

    def test_loadSmile(self):
        smi = SMILE_SMI
        sm = SmallMol(smi)
        n_atoms = sm.numAtoms
        self.assertEqual(n_atoms, SMILE_N_ATOMS, 'Atoms not correctly loaded. '
                                                  'Expected: {}; Now: {}'.format(SMILE_N_ATOMS, n_atoms))


    def test_getAtoms(self):
        smi = SMILE_SMI
        sm = SmallMol(smi)
        element_idx_1 = sm.get('element', 'idx 1')[0]
        neighbors_element_O = sm.get('neighbors', 'element O')[0]
        btypes_element_O = sm.get('bondtype', 'element O', convertType=False)[0]

        self.assertEqual(element_idx_1, PHENOL_ELEMENT_IDX_1, 'Element of the first atom does not correspond'
                                                              'Expect: {}; Now: {}'.format(element_idx_1, PHENOL_ELEMENT_IDX_1))
        self.assertListEqual(neighbors_element_O, PHENOL_ELEMENT_NEIGHBORS_OX,  'Neighbors atoms of the oxygen atom do not correspond'
                           'Expected: {}; Now: {}'.format(PHENOL_ELEMENT_NEIGHBORS_OX, neighbors_element_O))

        self.assertListEqual(btypes_element_O, PHENOL_BTYPES_OX, 'Bondtypes of the oxygen atom do not correspond:'
                                                                 'Expeected: {}; Now: {}'.format(btypes_element_O, PHENOL_BTYPES_OX))

    def test_isChiral(self):
        smi = CHIRAL_SMI
        sm = SmallMol(smi)
        ischiral, details = sm.isChiral(returnDetails=True)
        self.assertListEqual(details, CHIRAL_DETAILS, 'chiral atom does not match.'
                                                      'Expected: {}; Now: {}'.format(CHIRAL_DETAILS, details))

    def test_foundBond(self):
        smi = FOUNDBOND_SMI
        sm = SmallMol(smi)
        isbond_0_N = sm.foundBondBetween('idx 0', 'element N')
        isbond_0_1_single = sm.foundBondBetween('idx 0', 'idx 1', bondtype=1)
        isbond_0_1_double, _ = sm.foundBondBetween('idx 0', 'idx 1', bondtype=2)


        self.assertFalse(isbond_0_N, 'Bond between atom 0 and any nitrogens should not be present')
        self.assertFalse(isbond_0_1_single, 'Bond between atom 0 1 should not be single')
        self.assertTrue(isbond_0_1_double, 'Bond between atom 0 1 should  be double')


    def test_generateConformers(self):
        sm = SmallMol(self.benzamidine_mol2)
        current_conformer = sm.numFrames
        sm.generateConformers(num_confs=10, append=False)
        n_conformers = sm.numFrames

        self.assertGreater(n_conformers, current_conformer, 'The generation of conforemr should provide at least the '
                                                            'same amount of conformer')

    def test_writeGenerateAndWriteConformers(self):
        sm = SmallMol(self.benzamidine_mol2)
        sm.generateConformers(num_confs=10, append=False)
        tmpfname = os.path.join(NamedTemporaryFile().name, 'benzamidine.sdf')
        tmpdir = os.path.dirname(tmpfname)
        sm.write(tmpfname, merge=False)
        direxists = os.path.isdir(tmpdir)
        n_files = len(glob(os.path.join(tmpdir, '*.sdf')))
        self.assertTrue(direxists, 'The directory where to store the conformations where not created')
        self.assertGreater(n_files, 1, 'None conformations were written. At least one should be present')


    def test_removeGenerateConformer(self):
        molsmile = SMILE_SMI
        sm = SmallMol(molsmile)
        sm.generateConformers(num_confs=10, append=False)
        n_confs = sm.numFrames
        sm.dropFrames([0])
        n_confs_del = sm.numFrames
        sm.dropFrames()
        n_confs_zero = sm.numFrames

        self.assertEqual(n_confs_del, n_confs - 1, "The number of conformations after the deletion was not reduced of "
                                                   "exactly one unit")
        self.assertEqual(n_confs_zero, 0, "The number of conformations after the deletion was not reduced to 0")


    def test_convertToMolecule(self):
        from moleculekit.molecule import mol_equal 

        sm = SmallMol(self.benzamidine_mol2)
        mol = sm.toMolecule(formalcharges=False)

        assert mol_equal(sm, mol, exceptFields=['serial', 'name'], _logger=False)

    def test_convertFromMolecule(self):
        from moleculekit.molecule import mol_equal 

        mol = Molecule(self.benzamidine_mol2)
        sm = SmallMol(mol)

        assert mol_equal(sm, mol, exceptFields=['serial', 'name'], _logger=False)

    def test_getBonds(self):
        sm = SmallMol(self.benzamidine_mol2)

        self.assertListEqual(sm._bonds.tolist(), BENZAMIDINE_BOND_ATOMS, msg="The atoms in bonds are not the same of the reference")

        self.assertListEqual(sm._bondtype.tolist(), BENZAMIDINE_BONDTYPES, msg="The bonds type are not the same of the reference")

    @unittest.skipIf(rdkit.__version__ == '2019.03.4', "new rdkit versions break 2D depiction")
    def test_depict(self):
        import IPython
        refimg = os.path.join(self.dataDir, 'benzamidine.svg')

        sm = SmallMol(self.benzamidine_mol2)

        img_name = NamedTemporaryFile().name + '.svg'
        sm.depict(sketch=True, filename=img_name, atomlabels="%a%i%c")
        png_name = NamedTemporaryFile().name + '.png'
        sm.depict(sketch=True, filename=png_name, atomlabels="%a%i%c")
        noext_name = NamedTemporaryFile().name
        sm.depict(sketch=True, filename=noext_name, atomlabels="%a%i%c")
        sm.depict(sketch=False, optimize=True)
        _img = sm.depict(sketch=True, ipython=True)

        refimg_size = os.path.getsize(refimg)
        sm_img_size = os.path.getsize(img_name)

        self.assertIsInstance(_img, IPython.core.display.SVG, msg="The object is not an IPython image as expected")
        self.assertEqual(sm_img_size, refimg_size, msg="The svg image does not have the same size of the reference")

    def test_repr(self):
        sm = SmallMol(self.benzamidine_mol2)
        _ = str(sm)

    def test_toSMILES(self):
        sm = SmallMol(self.benzamidine_mol2)
        assert sm.toSMILES() == 'NC(=[NH2+])C1=CC=CC=C1', 'Failed with SMILES: {}'.format(sm.toSMILES())

    def test_toSMARTS(self):
        sm = SmallMol(self.benzamidine_mol2)
        assert sm.toSMARTS() == '[#6]1(:[#6H]:[#6H]:[#6H]:[#6H]:[#6H]:1)-[#6](=[#7H2+])-[#7H2]', 'Failed with SMARTS: {}'.format(sm.toSMARTS())

    def test_align(self):
        from moleculekit.util import rotationMatrix
        import numpy as np

        sm = SmallMol(self.benzamidine_mol2)
        mol = sm.toMolecule()
        mol.rotateBy(rotationMatrix([1, 0, 0], 3.14))
        sm.align(mol)
        
        assert (np.abs(sm._coords) - np.abs(mol.coords)).max()  # I need to do the abs of the coords since it's a symmetrical molecule

    # def test_getRCSBLigandByLigname(self):
    #     from moleculekit.smallmol.util import getRCSBLigandByLigname
    #     sm = getRCSBLigandByLigname('BEN')



if __name__ == '__main__':
    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from moleculekit.projections.projection import Projection
from moleculekit.projections.util import pp_calcDistances, pp_calcMinDistances
from moleculekit.util import ensurelist
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricDistance(Projection):
    """ Creates a MetricDistance object

    Parameters
    ----------
    sel1 : str
        Atom selection string for the first set of atoms.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    sel2 : str
        Atom selection string for the second set of atoms. If sel1 != sel2, it will calculate inter-set distances.
        If sel1 == sel2, it will calculate intra-set distances.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    groupsel1 : ['all','residue'], optional
        Group all atoms in `sel1` to the single minimum distance. Alternatively can calculate the minimum distance of a
        residue containing the atoms in sel1.
    groupsel2 : ['all','residue'], optional
        Same as groupsel1 but for `sel2`
    metric : ['distances','contacts'], optional
        Set to 'contacts' to calculate contacts instead of distances
    threshold : float, optional
        The threshold under which a distance is considered in contact. Units in Angstrom.
    pbc : bool, optional
        Set to false to disable distance calculations using periodic distances
    truncate : float, optional
        Set all distances larger than `truncate` to `truncate`. Units in Angstrom.
    update :
        Not functional yet

    Returns
    -------
    proj : MetricDistance object
    """

    def __init__(
        self,
        sel1,
        sel2,
        groupsel1=None,
        groupsel2=None,
        metric="distances",
        threshold=8,
        pbc=True,
        truncate=None,
    ):
        super().__init__()

        self.sel1 = sel1
        self.sel2 = sel2
        self.groupsel1 = groupsel1
        self.groupsel2 = groupsel2
        self.metric = metric
        self.threshold = threshold
        self.pbc = pbc
        self.truncate = truncate

    def _checkChains(self, mol, sel1, sel2):
        if np.array_equal(sel1, sel2):
            return
        if np.ndim(sel1) == 1:
            sel1 = sel1[None, :]
        if np.ndim(sel2) == 1:
            sel2 = sel2[None, :]
        sel1 = np.any(sel1, axis=0)
        sel2 = np.any(sel2, axis=0)
        chains_sel1 = mol.chain[sel1]
        chains_sel2 = mol.chain[sel2]
        if len(np.intersect1d(chains_sel1, chains_sel2)):
            logger.warning(
                "Atomselections sel1 and sel2 of MetricDistance contain atoms belonging to a common chain. "
                "Atoms within the same chain will not have periodic distances computed. "
                "Ensure that chains are properly defined in your topology file."
            )

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        sel1 = getMolProp("sel1")
        sel2 = getMolProp("sel2")
        if self.pbc:
            self._checkChains(mol, sel1, sel2)

        if np.ndim(sel1) == 1 and np.ndim(sel2) == 1:  # normal distances
            metric = pp_calcDistances(
                mol,
                sel1,
                sel2,
                self.metric,
                self.threshold,
                self.pbc,
                truncate=self.truncate,
            )
        else:  # minimum distances by groups
            metric = pp_calcMinDistances(
                mol,
                sel1,
                sel2,
                self.metric,
                self.threshold,
                pbc=self.pbc,
                truncate=self.truncate,
            )

        return metric

    def _calculateMolProp(self, mol, props="all"):
        props = ("sel1", "sel2") if props == "all" else props
        res = {}
        if "sel1" in props:
            res["sel1"] = self._processSelection(mol, self.sel1, self.groupsel1)
        if "sel2" in props:
            res["sel2"] = self._processSelection(mol, self.sel2, self.groupsel2)
        return res

    def _processSelection(self, mol, sel, groupsel):
        # If user passes simple string selections or 1D array of ints or bools
        if isinstance(sel, str) or (
            isinstance(sel, np.ndarray)
            and sel.ndim == 1
            and (np.issubdtype(sel.dtype, np.integer) or sel.dtype == bool)
        ):
            if groupsel is None:
                sel = mol.atomselect(sel)
            elif groupsel == "all":
                sel = self._processMultiSelections(mol, [sel])
            elif groupsel == "residue":
                sel = self._groupByResidue(mol, sel)
            else:
                raise NameError("Invalid groupsel argument")
        elif isinstance(sel, np.ndarray) or isinstance(
            sel, list
        ):  # If user passes his own sets of groups
            sel = self._processMultiSelections(mol, sel)
        else:
            raise RuntimeError(
                "Invalid atom selection. Either provide a string, a list of string, a 1D numpy array (int/bool) or a 2D numpy array for groups."
            )

        if np.sum(sel) == 0:
            raise NameError("Selection returned 0 atoms")
        return sel

    def _processMultiSelections(self, mol, sel):
        newsel = np.zeros((len(sel), mol.numAtoms), dtype=bool)
        for s in range(len(sel)):
            if isinstance(sel[s], str):
                newsel[s, :] = mol.atomselect(sel[s])
            elif isinstance(sel[s], np.ndarray) and (
                np.issubdtype(sel[s].dtype, np.integer) or sel[s].dtype == bool
            ):
                newsel[s, sel[s]] = True
            else:
                raise RuntimeError("Invalid selection provided for groups")
        return newsel

    def _groupByResidue(self, mol, sel):
        import pandas as pd

        idx = mol.atomselect(sel, indexes=True)
        df = pd.DataFrame({"a": mol.resid[idx]})
        gg = df.groupby(by=df.a).groups  # Grouping by same resids

        newsel = np.zeros((len(gg), mol.numAtoms), dtype=bool)
        for i, res in enumerate(sorted(gg)):
            newsel[
                i, idx[gg[res]]
            ] = True  # Setting the selected indexes to True which correspond to the same residue
        return newsel

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)
        sel1 = getMolProp("sel1")
        sel2 = getMolProp("sel2")

        if np.ndim(sel1) == 2:
            protatoms = []
            for i in range(sel1.shape[0]):
                protatoms.append(np.where(sel1[i, :] == True)[0])
        else:
            protatoms = np.where(sel1)[0]
        if np.ndim(sel2) == 2:
            ligatoms = []
            for i in range(sel2.shape[0]):
                ligatoms.append(np.where(sel2[i, :] == True)[0])
        else:
            ligatoms = np.where(sel2)[0]

        numatoms1 = len(protatoms)
        numatoms2 = len(ligatoms)

        from pandas import DataFrame

        types = []
        indexes = []
        description = []
        if np.array_equal(sel1, sel2):
            for i in range(numatoms1):
                for j in range(i + 1, numatoms1):
                    atm1 = protatoms[i]
                    atm2 = protatoms[j]
                    desc = "{} between {} {} {} and {} {} {}".format(
                        self.metric[:-1],
                        mol.resname[atm1],
                        mol.resid[atm1],
                        mol.name[atm1],
                        mol.resname[atm2],
                        mol.resid[atm2],
                        mol.name[atm2],
                    )
                    types += [self.metric[:-1]]
                    indexes += [[atm1, atm2]]
                    description += [desc]
        else:
            for i in range(numatoms1):
                for j in range(numatoms2):
                    atm1 = protatoms[i]
                    atm2 = ligatoms[j]
                    desc = "{} between {} {} {} and {} {} {}".format(
                        self.metric[:-1],
                        mol.resname[atm1],
                        mol.resid[atm1],
                        mol.name[atm1],
                        mol.resname[atm2],
                        mol.resid[atm2],
                        mol.name[atm2],
                    )
                    types += [self.metric[:-1]]
                    indexes += [[atm1, atm2]]
                    description += [desc]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


class MetricSelfDistance(MetricDistance):
    def __init__(
        self,
        sel,
        groupsel=None,
        metric="distances",
        threshold=8,
        pbc=True,
        truncate=None,
    ):
        """ Creates a MetricSelfDistance object

        Parameters
        ----------
        sel : str
            Atom selection string for which to calculate the self distance.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        groupsel : ['all','residue'], optional
            Group all atoms in `sel` to the single minimum distance. Alternatively can calculate the minimum distance
            of a residue containing the atoms in `sel`.
        metric : ['distances','contacts'], optional
            Set to 'contacts' to calculate contacts instead of distances
        threshold : float, optional
            The threshold under which a distance is considered in contact
        pbc : bool, optional
            Set to false to disable distance calculations using periodic distances
        truncate : float, optional
            Set all distances larger than `truncate` to `truncate`
        update :
            Not functional yet

        Returns
        -------
        proj : MetricDistance object
        """
        super().__init__(sel, sel, groupsel, groupsel, metric, threshold, pbc, truncate)


def contactVecToMatrix(vector, atomIndexes):
    from copy import deepcopy

    if np.ndim(vector) != 1:
        raise RuntimeError(
            "Please pass a 1D vector to the contactVecToMatrix function."
        )
    # Calculating the unique atom groups in the mapping
    uqAtomGroups = []
    atomIndexes = deepcopy(list(atomIndexes))
    for ax in atomIndexes:
        ax[0] = ensurelist(ax[0])
        ax[1] = ensurelist(ax[1])
        if ax[0] not in uqAtomGroups:
            uqAtomGroups.append(ax[0])
        if ax[1] not in uqAtomGroups:
            uqAtomGroups.append(ax[1])
    uqAtomGroups.sort(key=lambda x: x[0])  # Sort by first atom in each atom list
    num = len(uqAtomGroups)

    matrix = np.zeros((num, num), dtype=vector.dtype)
    mapping = np.ones((num, num), dtype=int) * -1
    for i in range(len(vector)):
        row = uqAtomGroups.index(atomIndexes[i][0])
        col = uqAtomGroups.index(atomIndexes[i][1])
        matrix[row, col] = vector[i]
        matrix[col, row] = vector[i]
        mapping[row, col] = i
        mapping[col, row] = i
    return matrix, mapping, uqAtomGroups


def reconstructContactMap(
    vector,
    mapping,
    truecontacts=None,
    plot=True,
    figsize=(7, 7),
    dpi=80,
    title=None,
    outfile=None,
    colors=None,
):
    """ Plots a given vector as a contact map

    Parameters
    ----------
    vector : np.ndarray or list
        A 1D vector of contacts
    mapping : pd.DataFrame
        A pandas DataFrame which describes the dimensions of the projection
    truecontacts : np.ndarray or list
        A 1D vector of true contacts
    plot : bool
        To plot or not to plot
    figsize : tuple
        The size of the final plot in inches
    dpi : int
        Dots per inch
    outfile : str
        Path of file in which to save the plot

    Returns
    -------
    cm : np.ndarray
        The input vector converted into a 2D numpy array

    Examples
    --------
    >>> reconstructContactMap(contacts, mapping)
    To use it with distances instead of contacts pass ones as the concat vector
    >>> reconstructContactMap(np.ones(dists.shape, dtype=bool), mapping, colors=dists)
    """
    from copy import deepcopy
    from matplotlib import cm as colormaps

    if np.ndim(vector) != 1:
        raise RuntimeError(
            "Please pass a 1D vector to the reconstructContactMap function."
        )

    if truecontacts is None:
        truecontacts = np.zeros(len(vector), dtype=bool)
    if len(vector) != len(mapping):
        raise RuntimeError("Vector and map length must match.")

    # Checking if contacts or distances exist in the data
    contactidx = mapping.type == "contact"
    if not np.any(contactidx):
        contactidx = mapping.type == "distance"
        if not np.any(contactidx):
            raise RuntimeError(
                "No contacts or distances found in the MetricData object. Check the `.map` property of the object for a description of your projection."
            )

    # Creating the 2D contact maps
    cm, newmapping, uqAtomGroups = contactVecToMatrix(vector, mapping.atomIndexes)
    cmtrue, _, _ = contactVecToMatrix(truecontacts, mapping.atomIndexes)
    num = len(uqAtomGroups)

    if plot:
        from matplotlib import pylab as plt

        f = plt.figure(figsize=figsize, dpi=dpi)
        plt.imshow(
            cmtrue / 2,
            interpolation="none",
            vmin=0,
            vmax=1,
            aspect="equal",
            cmap="Greys",
        )  # /2 to convert to gray from black

        rows, cols = np.where(cm)
        colorbar = False
        if colors is None:
            truecms = vector & truecontacts
            colors = np.array(["r"] * len(vector), dtype=object)
            colors[truecms] = "#ffff00"
        elif isinstance(colors, np.ndarray) and isinstance(colors[0], float):
            mpbl = colormaps.ScalarMappable(cmap=colormaps.jet)
            mpbl.set_array(colors)
            colors = mpbl.to_rgba(colors)
            colorbar = True
        if len(colors) == len(vector):
            colors = colors[newmapping[rows, cols]]

        plt.scatter(rows, cols, s=figsize[0] * 5, marker="o", c=colors, lw=0)
        if colorbar:
            plt.colorbar(mpbl)

        ax = f.axes[0]
        # Major ticks
        ax.set_xticks(np.arange(0, num, 1))
        ax.set_yticks(np.arange(0, num, 1))

        # Labels for major ticks
        ax.set_xticklabels([x[0] for x in uqAtomGroups])
        ax.set_yticklabels([x[0] for x in uqAtomGroups],)

        # Minor ticks
        ax.set_xticks(np.arange(-0.5, num, 1), minor=True)
        ax.set_yticks(np.arange(-0.5, num, 1), minor=True)
        plt.setp(ax.xaxis.get_majorticklabels(), rotation=90)

        # Gridlines based on minor ticks
        ax.grid(which="minor", color="#969696", linestyle="-", linewidth=1)
        ax.tick_params(axis="both", which="both", length=0)
        plt.xlim([-0.5, num - 0.5])
        plt.ylim([-0.5, num - 0.5])
        plt.xlabel("Atom index")
        plt.ylabel("Atom index")
        if title:
            plt.title(title)
        if outfile is not None:
            plt.savefig(outfile, dpi=dpi, bbox_inches="tight", pad_inches=0.2)
            plt.close()
        else:
            plt.show()
    return cm


import unittest
from moleculekit.home import home
import os


class _TestMetricDistance(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from os import path

        self.mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.pdb")
        )
        self.mol_skipped = self.mol.copy()

        self.mol.read(
            path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc")
        )
        self.mol_skipped.read(
            path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"),
            skip=10,
        )

    def test_contacts(self):
        metr = MetricDistance(
            "protein and name CA", "resname MOL and noh", metric="contacts", threshold=8
        )
        data = metr.project(self.mol)
        reffile = os.path.join(
            home(dataDir="test-projections"), "metricdistance", "distances.npy"
        )
        refdata = np.load(reffile) < 8

        assert np.allclose(data, refdata, atol=1e-3), "Contact calculation is broken"

    def test_atomselects(self):
        # String atomselection
        metr = MetricSelfDistance("protein and name CA", metric="contacts", threshold=8)
        data = metr.project(self.mol_skipped)
        mapping = metr.getMapping(self.mol_skipped)
        assert data.shape == (20, 38226) and mapping.shape == (38226, 3)

        # Integer index atomselection
        ca_atoms = self.mol_skipped.atomselect("protein and name CA", indexes=True)
        metr = MetricSelfDistance(ca_atoms, metric="contacts", threshold=8)
        data = metr.project(self.mol_skipped)
        mapping = metr.getMapping(self.mol_skipped)
        assert data.shape == (20, 38226) and mapping.shape == (38226, 3)

        # Boolean atomselection
        ca_atoms = self.mol_skipped.atomselect("protein and name CA")
        metr = MetricSelfDistance(ca_atoms, metric="contacts", threshold=8)
        data = metr.project(self.mol_skipped)
        mapping = metr.getMapping(self.mol_skipped)
        assert data.shape == (20, 38226) and mapping.shape == (38226, 3)

        # Manual group atomselection with integers
        ca_atoms = self.mol_skipped.atomselect("protein and name CA", indexes=True)
        metr = MetricSelfDistance([ca_atoms[0::3], ca_atoms[1::3], ca_atoms[2::3]],)
        data = metr.project(self.mol_skipped)
        mapping = metr.getMapping(self.mol_skipped)
        assert data.shape == (20, 3) and mapping.shape == (3, 3)

        # Manual group atomselection with booleans
        ca_atoms = self.mol_skipped.atomselect("protein and name CA", indexes=True)
        ca_bools1 = self.mol_skipped.atomselect(
            "index {}".format(" ".join(map(str, ca_atoms[0::3])))
        )
        ca_bools2 = self.mol_skipped.atomselect(
            "index {}".format(" ".join(map(str, ca_atoms[1::3])))
        )
        ca_bools3 = self.mol_skipped.atomselect(
            "index {}".format(" ".join(map(str, ca_atoms[2::3])))
        )
        metr = MetricSelfDistance([ca_bools1, ca_bools2, ca_bools3])
        newdata = metr.project(self.mol_skipped)
        mapping = metr.getMapping(self.mol_skipped)
        assert (
            newdata.shape == (20, 3)
            and mapping.shape == (3, 3)
            and np.array_equal(data, newdata)
        )

    def test_distances_trivial(self):
        from moleculekit.molecule import Molecule
        import numpy as np

        mol = Molecule().empty(3)
        mol.name[:] = "C"
        mol.element[:] = "C"
        mol.chain[:] = list(
            map(str, range(3))
        )  # If they are in the same chain, no wrapping is done for distances
        mol.coords = np.zeros(
            (3, 3, 2), dtype=np.float32
        )  # Make two frames so we check if the code works for nframes
        mol.coords[1, :, 0] = [3, 3, 3]
        mol.coords[2, :, 0] = [5, 5, 5]
        mol.coords[1, :, 1] = [7, 7, 7]
        mol.coords[2, :, 1] = [6, 6, 6]

        realdistances = np.linalg.norm(mol.coords[[1, 2], :, :], axis=1).T

        metr = MetricDistance("index 0", "index 1 2", metric="distances", pbc=False)
        data = metr.project(mol)
        assert np.allclose(
            data, realdistances
        ), "Trivial distance calculation is broken"

        # Test wrapped distances
        wrappedcoords = np.mod(mol.coords, 2)
        wrappedrealdistances = np.linalg.norm(wrappedcoords[[1, 2], :, :], axis=1).T

        mol.box = np.full((3, 2), 2, dtype=np.float32)  # Make box 2x2x2A large
        metr = MetricDistance("index 0", "index 1 2", metric="distances", pbc=True)
        data = metr.project(mol)
        assert np.allclose(
            data, wrappedrealdistances
        ), "Trivial wrapped distance calculation is broken"

        # Test min distances
        metr = MetricDistance(
            "index 0",
            "index 1 2",
            metric="distances",
            pbc=False,
            groupsel1="all",
            groupsel2="all",
        )
        data = metr.project(mol)
        assert np.allclose(
            data.flatten(), np.min(realdistances, axis=1)
        ), "Trivial distance calculation is broken"

        # Test ordering
        mol = Molecule().empty(4)
        mol.name[:] = "C"
        mol.element[:] = "C"
        mol.chain[:] = list(
            map(str, range(4))
        )  # If they are in the same chain, no wrapping is done for distances
        mol.coords = np.zeros(
            (4, 3, 2), dtype=np.float32
        )  # Make two frames so we check if the code works for nframes
        mol.coords[1, :, 0] = [1, 1, 1]
        mol.coords[2, :, 0] = [3, 3, 3]
        mol.coords[3, :, 0] = [5, 5, 5]
        mol.coords[1, :, 1] = [1, 1, 1]
        mol.coords[2, :, 1] = [7, 7, 7]
        mol.coords[3, :, 1] = [6, 6, 6]

        realdistances = np.linalg.norm(
            mol.coords[[2, 3], :, :] - mol.coords[0], axis=1
        ).T
        realdistances = np.hstack(
            (
                realdistances,
                np.linalg.norm(mol.coords[[2, 3], :, :] - mol.coords[1], axis=1).T,
            )
        )

        metr = MetricDistance("index 0 1", "index 2 3", metric="distances", pbc=False)
        data = metr.project(mol)
        assert np.allclose(
            data, realdistances
        ), "Trivial distance calculation has broken ordering"

    def test_distances(self):
        metr = MetricDistance(
            "protein and name CA", "resname MOL and noh", metric="distances"
        )
        data = metr.project(self.mol)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"), "metricdistance", "distances.npy"
            )
        )
        assert np.allclose(data, refdata, atol=1e-3), "Distance calculation is broken"

    def test_mindistances(self):
        metr = MetricDistance(
            "protein and noh",
            "resname MOL and noh",
            groupsel1="residue",
            groupsel2="all",
        )
        data = metr.project(self.mol)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"), "metricdistance", "mindistances.npy"
            )
        )
        assert np.allclose(
            data, refdata, atol=1e-3
        ), "Minimum distance calculation is broken"

    def test_mindistances_truncate(self):
        metr = MetricDistance(
            "protein and noh",
            "resname MOL and noh",
            groupsel1="residue",
            groupsel2="all",
            truncate=3,
        )
        data = metr.project(self.mol)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"), "metricdistance", "mindistances.npy"
            )
        )
        assert np.allclose(
            data, np.clip(refdata, 0, 3), atol=1e-3
        ), "Minimum distance calculation is broken"

    def test_selfmindistance_manual(self):
        metr = MetricDistance(
            "protein and resid 1 to 50 and noh",
            "protein and resid 1 to 50 and noh",
            groupsel1="residue",
            groupsel2="residue",
        )
        data = metr.project(self.mol)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"),
                "metricdistance",
                "selfmindistance.npy",
            )
        )
        assert np.allclose(data, refdata, atol=1e-3), "Manual self-distance is broken"

    def test_selfmindistance_auto(self):
        metr = MetricSelfDistance(
            "protein and resid 1 to 50 and noh", groupsel="residue"
        )
        data = metr.project(self.mol)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"),
                "metricdistance",
                "selfmindistance.npy",
            )
        )
        assert np.allclose(
            data, refdata, atol=1e-3
        ), "Automatic self-distance is broken"

    def test_mindistances_skip(self):
        metr = MetricSelfDistance(
            "protein and resid 1 to 50 and noh", groupsel="residue"
        )
        data = metr.project(self.mol_skipped)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"),
                "metricdistance",
                "selfmindistance.npy",
            )
        )
        assert np.allclose(
            data, refdata[::10, :], atol=1e-3
        ), "Minimum distance calculation with skipping is broken"

    def test_reconstruct_contact_map(self):
        from moleculekit.util import tempname
        from moleculekit.molecule import Molecule
        import matplotlib

        matplotlib.use("Agg")

        mol = Molecule("1yu8")

        metr = MetricSelfDistance("protein and name CA", metric="contacts", pbc=False)
        data = metr.project(mol)[0]
        mapping = metr.getMapping(mol)

        tmpfile = tempname(suffix=".svg")
        cm, newmapping, uqAtomGroups = contactVecToMatrix(data, mapping.atomIndexes)
        reconstructContactMap(data, mapping, outfile=tmpfile)
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"),
                "metricdistance",
                "contactvectomatrix.npz",
            )
        )

        assert np.array_equal(refdata["cm"], cm)
        assert np.array_equal(refdata["newmapping"], newmapping)
        assert np.array_equal(refdata["uqAtomGroups"], uqAtomGroups)

    def test_description(self):
        metr = MetricDistance("protein and noh", "resname MOL and noh", truncate=3)
        data = metr.project(self.mol)
        atomIndexes = metr.getMapping(self.mol).atomIndexes.values
        refdata = np.load(
            os.path.join(
                home(dataDir="test-projections"), "metricdistance", "description.npy"
            ),
            allow_pickle=True,
        )
        assert np.array_equal(refdata, atomIndexes)


if __name__ == "__main__":
    unittest.main(verbosity=2)
